from numpy import arange, zeros, array
from random import shuffle, uniform
from itertools import *
from math import ceil

from pylab import plot, show, title, figure

__all__ = ['LAST_BLOCK','SimulationResult','ReplayableEdgePolicy','single_simulate','simulate','differential_simulate']

LAST_BLOCK = -1

class SimulationResult:
	"""
	A class for representing the result of a tokenized simulation	
	"""
		
	def __init__(self,nr=None,i2n=None,n2i=None,er=None,i2e=None,e2i=None,name=None):
		self.node_result = nr
		self.idx2node = i2n
		self.nodename2idx = n2i
		
		self.edge_result = er
		self.idx2edge = i2e
		self.edge2idx = e2i 
		
		self.name = name
		
	def plot_node(self,node,new=True,style=None,line_width=None):
		if new:
			figure()
		
		if line_width is None:
			line_width = 1
		
		if style:
			plot(self.node_result[:,self.nodename2idx[node]],style,linewidth=line_width)
		else:
			plot(self.node_result[:,self.nodename2idx[node]],linewidth=line_width)
			
	def node_value(self,node,index):
		return self.node_result[index,self.nodename2idx[node]]
		
	def edge_weight(self,edge,tstart,tend=None):
		"""
		Return the edge weight over a time block interval.  If tend is not specified, then the
		interval is assumed to be the single time block, tstart.  If tend is LAST_BLOCK, then the 
		interval is from tstart to the final time block.
		"""
		if tend is None:
			return self.edge_result[tstart,self.edge2idx[edge]]
		elif tend == -1:
			tend = len(self.edge_result)
		
		sum = 0
		eidx = self.edge2idx[edge]
		for i in range(tstart,tend):
			sum += self.edge_result[i,eidx]
			
		return sum
		
	###
	# MarkingSeries methods
	###
	def num_molecules(self):
		return len(self.nodename2idx)
		
	def molecule_names(self):
		return self.nodename2idx.keys()
		
	def num_markings(self):
		return len(self.node_result[:,1])
		
	def get_molecule_value(self,mname,idx):
		return self.node_result[idx,self.nodename2idx[mname]]
		
	def get_molecule_series(self,mname):
		return self.node_result[:,self.nodename2idx[mname]]
		
	def get_name(self):
		return self.name
		
	def set_name(self,name):
		self.name = name
		
	def marking_type(self):
		return 'series'
		

# Edge evaluation policies
class ReplayableEdgePolicy:
	"""
	This edge policy supports replaying the a previously used policy.  After a set of
	weightings and orderings have been generated, future runs will use these same parameters.
	"""
	def __init__(self,edges): 
		self.num_edges = len(edges)
		self.edge_orderings = []
		self.edge_weightings = []
	
		self.edge_list = list(edges)

	def generate_ordering(self,tblock):
		"""
		Generate an ordering for block tblock.  If an ordering has been generated for
		this block before, then the previously generated ordering is used.  Otherwise,
		a new ordering is generated.
		"""
		if len(self.edge_orderings) > tblock:
			return self.edge_orderings[tblock]
		else:
			eordering = list(self.edge_list)
			shuffle(eordering)
			self.edge_orderings.append(eordering)
			return eordering
		
	def generate_weighting(self,tblock):
		"""
		Generate a weighting for block tblock.  If a weighting has been generated for
		this block before, then the previously generated weighting is used.  Otherwise,
		a new weighting is generated.
		"""
		if len(self.edge_weightings) > tblock:
			return self.edge_weightings[tblock]
		else:
			weighting = [ uniform(0,1) for i in range(0,self.num_edges) ]
			self.edge_weightings.append(weighting)
			return weighting			

# Simulation functions
def single_simulate(network,
					   	hi,low,
						num_blocks,
						edge_policy,
						high_val=10,
						prior_result=None,initial_marking=None):
	""" 
	Run the tokenized simulator on the specified signaling network.
	
	hi and low are both dictionaries mapping network node names (these are the labels of the
	nodes themselves) to number pairs (x,y) which specify the time block at which the node 
	becomes high/low (x) and the time block beyond which the node is no longer high/low (y).
	
	the hi dictionary can also map to triples (x,y,h) where h is the number of tokens assigned
	to the associated protein.  This number is only used if x == 0, otherwise, when the 
	protein becomes high it takes on its current token count (and this count never decreases).
	
	If the prior_result is specified, then the node and edge indices will be used from 
	this prior result to facilitate comaprison between it and the result of this
	simulation.
	
	The return value is a SimulationResult object.
	"""
	num_nodes = len(network.V)
	num_edges = len(network.E)

	# Allocate room for the results
	node_result = zeros((num_blocks,num_nodes))
	edge_result = zeros((num_blocks,num_edges))
	
	# Initialize node and edge lookups
	if prior_result:
		# sanity check
		if num_nodes != len(prior_result.idx2node) or num_edges != len(prior_result.idx2edge):
			raise Exception, "Network is inconsistent with prior result specified."
	
		nodename2idx = prior_result.nodename2idx
		idx2node = prior_result.idx2node

		edge2idx = prior_result.edge2idx
		idx2edge = prior_result.idx2edge
			
	else:
		# allocate node indices
		nodename2idx = {}
		idx2node = []
		
		for node in network.N:
			nodename2idx[node.label] = len(idx2node)
			idx2node.append(node)
	
		# allocate edge indices
		edge2idx = {}
		idx2edge = []
		
		for edge in network.E:
			edge2idx[edge] = len(idx2edge)
			idx2edge.append(edge)
	
	# determine the high value for each high protein
	hi_values = {}
	for k,v in hi.items():
		if len(v) == 2:
			hi_values[k] = high_val
		elif len(v) == 3:
			hi_values[k] = v[2]
		else:
			raise Exception, '2 or 3 values expected for each high protein'
	
	# Initial token assignment is all zeros, so nothing to do
	weighting = edge_policy.generate_weighting(0)
	ordering = edge_policy.generate_ordering(0)
	
	first_ta = node_result[0,:]
	for n in [name for name, value in hi.items() if value[0] == 0]:
		first_ta[nodename2idx[n]] = hi_values[n]
		
	# if the initial marking was given, apply it here
	if initial_marking:
		for n, value in initial_marking.items():
			first_ta[nodename2idx[n]] = value
	
	# run the simulation
	for tblock in arange(1,num_blocks):
		weighting = edge_policy.generate_weighting(tblock)
		ordering = edge_policy.generate_ordering(tblock)
		
		last_ta = node_result[tblock-1,:]
		node_result[tblock,:] = array(last_ta)
		this_ta = node_result[tblock,:]
		
		# Evaluate each edge in the ordering specified
		for edge, weight in izip(ordering,weighting):
			source = network.source(edge)
			sidx = nodename2idx[source.label]
			target = network.target(edge)
			tidx = nodename2idx[target.label]
			
			ntoks = ceil(weight * this_ta[sidx])
			
			if edge.head == '>':
				if target.label not in low or low[target.label][0] > tblock or low[target.label][1] < tblock:
					this_ta[tidx] += ntoks
			else: # edge is inhibiting edge.head == '|'
				if target.label not in hi or hi[target.label][0] > tblock or hi[target.label][1] < tblock:
					if ntoks > this_ta[tidx]: ntoks = this_ta[tidx] 
					this_ta[tidx] -= ntoks
			
			# update edge token count for this edge
			edge_result[tblock,edge2idx[edge]] = ntoks;
			
	return SimulationResult(node_result, idx2node, nodename2idx, edge_result, idx2edge, edge2idx)


def simulate(network, hi, low,
				num_runs, num_blocks,
				high_val=10,
				prior_result=None,initial_marking=None):
	"""
	Run the tokenized simulator num_runs times on the network with the settings specified.
	
	The SimulationResult returned is the average of the results.
	"""
	
	tot_result = prior_result
	
	for run in range(0,num_runs):
		ep = ReplayableEdgePolicy(network.E())
		result = single_simulate(network,hi,low,num_blocks,ep,high_val,tot_result,initial_marking)
		
		if tot_result is None:
			tot_result = result
		else:
			tot_result.node_result += result.node_result
			tot_result.edge_result += result.edge_result
		
	tot_result.node_result /= float(num_runs)
	tot_result.edge_result /= float(num_runs)
	
	return tot_result
		
def differential_simulate(network, 
					 hi1, low1, hi2, low2,						
					 num_runs, num_blocks, high_val=10,
					 prior_result=None,imarking1=None,imarking2=None):
	"""
	Run the simulator under two different conditions and compare the results.
	
	The result returned is the difference in the network's response to the two
	conditions (COND2 - COND1).
	"""
	
	tot_result = None
	
	for run in range(0,num_runs):
		ep = ReplayableEdgePolicy(network.E())
		result1 = single_simulate(network,hi1,low1,num_blocks,ep,high_val,tot_result,imarking1)
		result2 = single_simulate(network,hi2,low2,num_blocks,ep,high_val,result1,imarking2)
		
		if tot_result is None:
			tot_result = result2
			tot_result.node_result -= result1.node_result
			tot_result.edge_result -= result1.edge_result
		else:
			tot_result.node_result += (result2.node_result - result1.node_result)
			tot_result.edge_result += (result2.edge_result - result1.edge_result)		
		
	tot_result.node_result /= float(num_runs)
	tot_result.edge_result /= float(num_runs)
	
	return tot_result 
