import Tkinter
import tkFileDialog

from pathwayoracle.tkui.main import MainWindow

from marking import MarkingManagerController
from path import PathSetInspectorController

import pathwayoracle.io.marking as marking
import pathwayoracle.io.ncf as ncf
import pathwayoracle.io.path as path

import pathwayoracle.algorithm.tokenized_simulator as tok_sim
import pathwayoracle.algorithm.path_finder as pf

import pylab

class MainController:
	"""
	This is the main controller for the PathwayOracle application.
	"""
	
	def start(self, root):
	
		self.root = root
	
		# prompt for the network to analyze
		fname = tkFileDialog.askopenfilename(parent=self.root, title="Select a network file to analyze")
		
		if len(fname) == 0:
			print 'No network file selected'
			return
		
		# open the file
		self.network = ncf.read(fname)
		
		self.name2node = {}
		for n in self.network.N:
			self.name2node[n.label] = n
		
		# create the main window
		self.main_window = MainWindow(self, self.root, self.network)
		
		# create the marking manager
		self.mmanager = MarkingManagerController(self)
		
		# fields for keeping track of graphics
		self.path_count = {}
		
	def run_simulation(self,mname, nruns, nblocks, reorder, high_mols, low_mols):
		"""
		Run a tokenized simulation
		"""
		# sanity check to make sure that all molecules names exist
		mol_names = set([n.label for n in self.network.N])
		for h,v in high_mols:
			if h not in mol_names:
				raise Exception,'Molecule %s does not exist' % h
				
		for n in low_mols:
			if n not in mol_names:
				raise Exception,'Molecule %s does not exist' % n
			
		# make high and low dictionaries
		hi = dict([(h,(0,nblocks,v)) for h, v in high_mols])
		#hi = dict(zip(high_mols, [(0,nblocks)] * len(high_mols)))
		low = dict(zip(low_mols, [(0,nblocks)] * len(low_mols)))
		
		# run the simulator
		result = tok_sim.simulate(self.network, hi, low, nruns, nblocks)
		
		result.set_name(mname)
		
		# add the result to the marking manager
		self.mmanager.add_marking_series(result)
		
		# show the inspector for this series
		self.mmanager.inspect_series(result)
		
	def toggle_mm_visibility(self):
		"""
		Toggle the visibility of the marking manager
		"""
		self.mmanager.set_visible(not self.mmanager.get_visible())
	
	def run_diff_sim(self, mname, nruns, nblocks, reorder, hi_list1, low_mols1, hi_list2, low_mols2):
		"""
		Run a differential tokenized simulation
		"""
		hi1 = dict([(h,(0,nblocks,v)) for h, v in hi_list1])
		low1 = dict(zip(low_mols1, [(0,nblocks)] * len(low_mols1)))
		
		hi2 = dict([(h,(0,nblocks,v)) for h, v in hi_list2])
		low2 = dict(zip(low_mols2, [(0,nblocks)] * len(low_mols2)))		
	
		result = tok_sim.differential_simulate(self.network, hi1, low1, hi2, low2, nruns, nblocks)
		
		result.set_name(mname)
		
		# add the result to the marking manager
		self.mmanager.add_marking_series(result)
		
		# show the inspector for the series
		self.mmanager.inspect_series(result)

	def open_marking_group(self,fname):
		mgroup = marking.read_group(fname)
		self.mmanager.add_marking_group(mgroup)
		
		self.mmanager.inspect_group(mgroup)
		
	def open_marking_series(self,fname):
		mseries = marking.read_series(fname)
		self.mmanager.add_marking_series(mseries)
		
		self.mmanager.inspect_series(mseries)
		
	def save_network(self,fname):
		print 'Network',self.network
		ncf.write(self.network,fname)
		
	def compute_paths(self,src_name,tgt_name):
		
		src = None
		tgt = None
		for n in self.network.V:
			if n.label == src_name:
				src = n
			elif n.label == tgt_name:
				tgt = n
				
			if src and tgt:
				break
		
		paths = pf.find_paths(self.network,src,tgt)
		
		path_info = [(p,pf.classify_path(p)) for p in paths]
		
		PathSetInspectorController(self,path_info,self.network)
		
	def open_path_set(self,fname):
		paths = path.read(fname,self.network)
		
		path_info = [(p,pf.classify_path(p)) for p in paths]

		PathSetInspectorController(self,path_info,self.network)
		
	def color_network(self,molecule_names,molecule_colors):
		"""
		Color the network using the molcule names and colors specified.  Any molecules not in the
		network will not be colored - no error will be raised.
		"""
		for m,c in zip(molecule_names,molecule_colors):
			if m in self.name2node:
				self.main_window.color(self.name2node[m],c)
		
	def uncolor_network(self):
		"""
		Remove the colors that were applied to the network by any previous calls to color_network(...)
		"""
		self.main_window.reset_node_colors()
		
	def highlight_path(self,path,highlight=True):
		
		if highlight:
			self.main_window.set_width(path,3)
			
			for e in path:
				if e in self.path_count:
					self.path_count[e]+=1
				else:
					self.path_count[e] = 1
		else:
			for e in path:
				if e not in self.path_count:
					continue
				elif self.path_count[e] == 1:
					self.main_window.set_width(e,1)
					del self.path_count[e]
				else:
					self.path_count[e]-=1
		
def run():	
	
	# turn off interactive python
	pylab.ioff()
	
	# create the Tk context
	root = Tkinter.Tk()

	# run the main controller
	mc = MainController()
	mc.start(root)

	Tkinter.mainloop()