from pyweaver.graph import DiGraph

"""
This module provides functions that can read and write signaling networks from the Network 
Connectivity Format (.ncf).
"""

class SignalingNode:
	"""
	This class is a signaling node for use in the DiGraph.  It makes use of the rendering hint
	.label to specify the name of the node.  This class is implemented such that nodes stored in
	a hashable container can be retrieved by its string name or the object itself.
	"""

	def __init__(self,name):
		self.label = name
		
	def __eq__(self,val):
		return id(self) == id(val) or val == self.label
		
	def __hash__(self):
		return self.label.__hash__()
		
class SignalingEdge:
	"""
	This class represents a signaling edge.  It makes use of the .head rendering hint to represent
	the effect of the signaling edge.  a head of ">" indicates an activating edge, "|" indicates an
	inhibiting edge.
	
	"""
	def __init__(self,effect):
		self.head = effect;
		
def read(filename):
	"""
	Read a network from the Network Connnectivity Formatted file specified.
	"""
	G = DiGraph()
	G.name = filename
	
	# Read connectivity
	nodes = {}
	edges = {}
	
	line = None
	infile = open(filename,"r")
	while infile:
		line = infile.readline()

		# if the layout section has begun
		if line.startswith('BEGIN LAYOUT'):
			break
		
		# if the file ended
		if len(line) == 0:
			break
		
		# if a comment was inserted or if the line doesn't have any content
		if line.startswith('#') or line.isspace():
			continue
		
		data = line.split()
		
		if len(data) != 3:
			raise Exception, 'Line ' + line + ' is not a valid interaction'
			
		src_mol,effect,tgt_mol = data
		
		# fetch the nodes
		if src_mol not in nodes:
			nodes[src_mol] = SignalingNode(src_mol)
			
		if tgt_mol not in nodes:
			nodes[tgt_mol] = SignalingNode(tgt_mol)
		
		src = nodes[src_mol]
		tgt = nodes[tgt_mol]
		
		# make the edge
		if effect == '->':
			head = '>'
		elif effect == '-|':
			head = '|'
		else:
			raise Exception, 'Unrecognized effect: ' + effect
		
		edge = SignalingEdge(head)
		edges[' '.join([src_mol,tgt_mol])] = edge
		
		G.E.add(src,tgt,edge)
	
	# Read layout
	if line.startswith('BEGIN LAYOUT'):
		while infile:
			line = infile.readline()
			
			if len(line) == 0:
				break
			
			if line.startswith('#') or line.isspace():
				continue
				
			data = line.split()
			
			if data[0] == 'NODE':
				node = nodes[data[1]]
				node.x = float(data[2])
				node.y = float(data[3])
			elif data[0] == 'EDGE':
				eid = ' '.join([data[1],data[2]])
				
				if eid in edges:
					edge = edges[eid]
					waypoints = []
					for i in range(3,len(data)):
						if i % 2 != 0:
							waypoints.append((float(data[i]),float(data[i+1])))
	
					edge.waypoints = waypoints
			else:
				raise Exception, 'Unrecognized layout element: ' + data[0]
	
	# done.
	infile.close()
	
	return G
	
def write(G,filename):
	"""
	Write the specified network to the filename indicated in the Network Connectivity Format.
	"""
	outfile = open(filename,'w')
	
	# write all edges
	for e in G.E:
		src = G.source(e).label
		tgt = G.target(e).label
		effect = '-' + e.head
		outfile.write(' '.join([src,effect,tgt]) + '\n')
		
	# write all layouts
	outfile.write('BEGIN LAYOUT\n')
	for n in G.V:
		if hasattr(n,'x') and hasattr(n,'y'):
			outfile.write('NODE ' + ' '.join([n.label,str(n.x),str(n.y)]) + '\n')
	
	for e in G.E:
		src = G.source(e).label
		tgt = G.target(e).label
		if hasattr(e,'waypoints'):
			outfile.write('EDGE ' + src + ' ' + tgt + ' ')
			for w in e.waypoints:
				outfile.write(str(w[0]) + ' ' + str(w[1]) + ' ')
			outfile.write('\n')
			
	outfile.close()
	