
"""
This module defines markings of different forms.  These classes are models that can be used
for implementing other classes with the same interface.  Classes that are markings of some
sort may either extend the appropriate class or simply implement the same functions
(all of which are required).
"""

class Marking:
	
	def __init__(self,molecule2value,name=None):
		self.molecule2value = molecule2value
		self.name = name
	
	def get_name(self):
		return self.name
	
	def num_molecules(self):
		return len(self.molecule2value)
	
	def molecule_names(self):
		"""
		Return the list of molecule names that have values in this series.
		"""
		return self.molecule2value.keys()
		
	def get_molecule_value(self,mname):
		return self.molecule2value[mname]
		
	def marking_type(self):
		"""
		This function returns either "marking", "series", or "group".
		"""
		return "marking"
	

class MarkingSeries:
	"""
	This class provides methods for accessing data about a series of marking.
	"""
	
	def __init__(self,molecule2series,num_markings,name=None):
		"""
		molecule2series is the mapping of molecule name to its series data.
		num_markings is the number of markings recorded for each molecule.
		"""
		self.molecule2series = molecule2series
		self.num_markings = num_markings
		self.name = name
	
	def get_name(self):
		return self.name
	
	def num_molecules(self):
		return len(self.molecule2series)
	
	def molecule_names(self):
		"""
		Return the list of molecule names that have values in this series.
		"""
		return self.molecule2series.keys()
	
	def num_markings(self):
		return len(self.num_markings)
	
	def get_molecule_value(self,mname,idx):
		return self.molecule2series[mname][idx]
		
	def get_molecule_series(self,mname):
		"""
		Return a list/vector of values that correspond to the values for this molecule in the
		series.
		"""
		return self.molecule2series[mname]

	def marking_type(self):
		return "series"
		
class MarkingGroup:
	"""
	This class provides methods for accessing data about several markings.
	"""
	
	def __init__(self,group2values,molecule2idx,name=None):
		"""
		molecule2series is the mapping of molecule name to its series data.
		num_markings is the number of markings recorded for each molecule.
		"""
		self.molecule2idx = molecule2idx
		self.group2values = group2values
		self.name = name
	
	def get_name(self):
		return self.name
	
	def num_groups(self):
		return len(self.group2values)
	
	def group_names(self):
		return self.group2values.keys()
	
	def num_molecules(self):
		return len(self.molecule2idx)
	
	def molecule_names(self):
		"""
		Return the list of molecule names that have values in this series.
		"""
		return self.molecule2idx.keys()
	
	def get_molecule_value(self,mname,gname):
		return self.group2values[gname][self.molecule2idx[mname]]
		
	def marking_type(self):
		return "group"		