
"""
This module provides a heatmap Tkinter widget.
"""

import Tkinter as tk
from PIL import ImageFont, ImageTk, Image, ImageDraw

import pyweaver.visual.colors as pyg_colors

X_GUTTER = 10
Y_GUTTER = 10
GRID_GUTTER = 5

UNSELECTED_COLOR = "black"
SELECTED_COLOR = "blue"

class Heatmap(tk.Frame):
	"""
	The Heatmap widget is a Tkinter frame that displays the data as a grid of colors.  Row names are
	drawn to the left of each row.  Column names are drawn vertically above each column.  If the 
	displayed portion of the frame does not make the entire heatmap visible, then scroll bars are
	enabled and allow the user to pan across the entire heatmap display.
	
	Rows can be selected.  Objects that register as row selection listeners (via the add_row_listener
	method), will be notified when rows are selected or deselected via the methods:
	
	  - row_selected(heatmap,rnum)
	  
	  - row_deselected(heatmap,rnum)
	  
	At present only one row may be selected at a time.
	"""
	
	def __init__(self,root,data,row_names,col_names):
		tk.Frame.__init__(self,root)
		
		self.data = data
		self.row_names = row_names
		self.col_names = col_names
		
		# row selection stuff
		self.row_listeners = []
		self.selected_row = None
		self.selected_row_tid = None
		
		# build the ui
		self.build_ui()
		
	def build_ui(self):
		
		# build the heatmap canvas
		self.canvas = tk.Canvas(self)
		
		# create scroll bars
		self.scrollX = tk.Scrollbar(self, orient=tk.HORIZONTAL)
		self.scrollY = tk.Scrollbar(self, orient=tk.VERTICAL)
		
		# link the scroll bars and canvas
		self.canvas['xscrollcommand'] = self.scrollX.set
		self.canvas['yscrollcommand'] = self.scrollY.set
		self.scrollX['command'] = self.canvas.xview
		self.scrollY['command'] = self.canvas.yview

		# layout
		self.scrollX.pack(side=tk.BOTTOM, fill=tk.X)
		self.scrollY.pack(side=tk.RIGHT, fill=tk.Y)
		self.canvas.pack(side=tk.TOP,fill=tk.BOTH, expand=1)
		
		###
		# populate the canvas with the heatmap data
		self.canvas.bind("<Button-1>",lambda x: self.row_selected(x))
		# create row names
		self.tid_info = {}
		self.row_txtids = []
		max_rname_width = 0
		max_rname_height = 0
		
		i = 0
		for n in self.row_names:
			tid = self.canvas.create_text((0,0), text=n, fill=UNSELECTED_COLOR,anchor=tk.NW)
			
			self.row_txtids.append(tid)
			self.tid_info[tid] = i
			
			bounds = self.canvas.bbox(tid)
			width = bounds[2] - bounds[0]
			height = bounds[3] - bounds[1]
			
			if width > max_rname_width:
				max_rname_width = width
				
			if height > max_rname_height:
				max_rname_height = height
			
			i += 1
			
		# create col names
		self.vertical_images = []
		col_font = ImageFont.load_default()
		self.col_imgids = []
		max_cname_width = 0
		max_cname_height = 0
		
		for n in self.col_names:
			width,height = col_font.getsize(n)
			img = Image.new("1",(width,height),"white")
			pen = ImageDraw.Draw(img)
			
			pen.text( (0,0), n, font=col_font)
			rotated = img.rotate(-90)
			width,height = height,width
			
			gif = ImageTk.PhotoImage(rotated)
			self.vertical_images.append(gif)
			
			imgid = self.canvas.create_image(0,0,image=gif,anchor=tk.NW)
			self.col_imgids.append(imgid)
			bounds = self.canvas.bbox(imgid)
			width = bounds[2] - bounds[0]
			height = bounds[3] - bounds[1]
		
			if width > max_cname_width:
				max_cname_width = width
				
			if height > max_cname_height:
				max_cname_height = height
		
		# layout row names
		right_margin = X_GUTTER + max_rname_width
		y = Y_GUTTER + max_cname_height + GRID_GUTTER
		for tid in self.row_txtids:
			bounds = self.canvas.bbox(tid)
			width = bounds[2] - bounds[0]
			height = bounds[3] - bounds[1]
			
			x = right_margin - width
			self.canvas.coords(tid,(x,y))
			
			y += max_rname_height
		
		# layout col names
		top_margin = Y_GUTTER + max_cname_height
		x = X_GUTTER + max_rname_width + GRID_GUTTER
		for imgid in self.col_imgids:
			bounds = self.canvas.bbox(imgid)
			width = bounds[2] - bounds[0]
			height = bounds[3] - bounds[1]
			
			y = top_margin - height
			self.canvas.coords(imgid,(x,y))
			
			x += max_cname_width
			
		# find the max value in the data
		max_val = 0
		min_val = 0
		for row in self.data:
			for value in row:
				if value > max_val:
					max_val = value
					
				if value < min_val:
					min_val = value
		
		max_val = float(max_val)
		min_val = float(min_val)
		
		# Render the data
		self.row_rectids = []
		y = Y_GUTTER + max_cname_height + GRID_GUTTER
		
		for row in self.data:
			rectids = []
			x = X_GUTTER + max_rname_width + GRID_GUTTER
			for value in row:
				# select the color
				color = (0,0,0)
				if value > 0:
					color = (0, int(float(value) / max_val * 255.0) , 0)
				elif value < 0:
					color = (int(float(value) / min_val * 255.0) , 0, 0)
				
				# convert the color to a string
				color = "#%02x%02x%02x" % color
				
				rid = self.canvas.create_rectangle((x,y,x+max_cname_width,y+max_rname_height),fill=color)
				rectids.append(rid)
				
				x += max_cname_width
			self.row_rectids.append(rectids)
			y += max_rname_height
			
		# set the scrollable region
		total_width = X_GUTTER + max_rname_width + GRID_GUTTER + (max_cname_width * len(self.col_names))
		total_height = Y_GUTTER + max_cname_height + GRID_GUTTER + (max_rname_height * len(self.row_names))
		self.canvas.config(scrollregion=(0,0,total_width,total_height))		
		
	def add_row_listener(self,listener):
		self.row_listeners.append(listener)
		
	def rm_row_listener(self,listener):
		self.row_listeners.remove(listener)
		
	def row_selected(self,event):
		"""
		The callback used to process row selection.
		"""
		tids = self.canvas.find_withtag(tk.CURRENT)
		
		# deselect the last row
		if self.selected_row_tid is not None:
			sr = self.selected_row
			self.selected_row = None

			self.canvas.itemconfigure(self.selected_row_tid,fill=UNSELECTED_COLOR)

			for l in self.row_listeners:
				l.row_deselected(self,sr)
		
		# this row is selected
		if len(tids) > 0:
			tid = tids[0]
			self.selected_row_tid = tid
			self.selected_row = self.tid_info[tid]
			
			self.canvas.itemconfigure(self.selected_row_tid,fill=SELECTED_COLOR)
			
			for l in self.row_listeners:
				l.row_selected(self,self.selected_row)
			
	def selected_row():
		"""
		Return the index of the row that is currently selected.  None is returned if no row is
		currently selected.
		"""
		return self.selected_row

	def column_names(self):
		return self.col_names
		
	def row_colors(self,ridx,style='tkrgb'):
		"""
		Return the colors for individual values in the specified row.
		The style argument selects what format the color values will be returned in.  Default is the
		RGB string format used by Tkinter.  Valid values are:
		  
		  - 'tkrgb' - the default RBG string format used by Tkinter
		  - 'pygraphrgb' - the RBG tuple format used by PyGraph
		"""	
		colors = None
		
		if style == 'tkrgb':
			colors = [self.canvas.itemcget(rid,'fill') for rid in self.row_rectids[ridx]]
		elif style == 'pygraphrgb':
			colors = []
			for color_str in [self.canvas.itemcget(rid,'fill') for rid in self.row_rectids[ridx]]:
				colors.append(pyg_colors.tkrgb2color(color_str))
		else:
			raise Exception,'Unsupported color style: ' + str(style)
		
		return colors
		
