"""
This module provides implementations of various measures of centrality.
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

import types

from graph import Graph, DiGraph
import shortestpath as sp
import heapq

__all__ = ['betweenness','brandes_betweenness','pairwise_disconnectivity']

#####
# Betweenness functions

def brandes_betweenness(G,normalize=True,edge_weight_fxn=None):
	"""
	Compute the betweenness centrality for nodes in G: 
	the fraction of number of shortest paths that pass through each node. 
 
	The keyword normalize (default=True) specifies whether the  
	betweenness values are normalized by b=b/(n-1)(n-2) where 
	n is the number of nodes in G. 
 
	The keyword edge_weight_fxn specifies a function that returns the weight
	of the edges.  By default, the function assumes unit weight.

	The algorithm is from 
	Ulrik Brandes, 
	A Faster Algorithm for Betweenness Centrality. 
	Journal of Mathematical Sociology 25(2):163-177, 2001. 
	http://www.inf.uni-konstanz.de/algo/publications/b-fabc-01.pdf 

	This implementation was adapted from a version in NetworkX (http://networkx.lanl.gov)
	"""
	
	neighbors = {}
	for v in G.V:
		neighbors[v] = G.neighbors(v)
	
	betweenness=dict.fromkeys(G.V,0.0) # b[v]=0 for v in G 
	for s in G.V: 
		S=[] 
		P={} 
		for v in G.V: 
			P[v]=[] 
		sigma=dict.fromkeys(G.V,0)	# sigma[v]=0 for v in G 
		D={} 
		sigma[s]=1 
		if edge_weight_fxn is None:  # use BFS 
			D[s]=0 
			Q=[s] 
			while Q:   # use BFS to find shortest paths 
				v=Q.pop(0) 
				S.append(v) 
				for w in neighbors[v]: #G.neighbors(v): 
					if w not in D: 
						Q.append(w) 
						D[w]=D[v]+1 
					if D[w]==D[v]+1:   # this is a shortest path, count paths 
						sigma[w]=sigma[w]+sigma[v] 
						P[w].append(v) # predecessors  
		else:  # use Dijkstra's algorithm for shortest paths, 
			   # modified from Eppstein 
			push=heapq.heappush 
			pop=heapq.heappop 
			seen = {s:0}  
			Q=[]   # use Q as heap with (distance,node id) tuples 
			push(Q,(0,s,s)) 
			while Q:	
				(dist,pred,v)=pop(Q) 
				if v in D: 
					continue # already searched this node. 
				sigma[v]=sigma[v]+sigma[pred] # count paths 
				S.append(v) 
				D[v] = seen[v] 
				for w in neighbors[v]: #G.neighbors(v): 
					vw_dist = D[v] + edge_weight_fxn(G.E.get(v,w)) 
					if w not in D and (w not in seen or vw_dist < seen[w]): 
						seen[w] = vw_dist 
						push(Q,(vw_dist,v,w)) 
						P[w]=[v] 
					elif vw_dist==seen[w]:  # handle equal paths 
						sigma[w]=sigma[w]+sigma[v] 
						P[w].append(v) 
 
 
		delta=dict.fromkeys(G.V,0)  
		while S: 
			w=S.pop() 
			for v in P[w]: 
				delta[v]=delta[v]+(float(sigma[v])/float(sigma[w]))*(1.0+delta[w]) 
			if w != s: 
				betweenness[w]=betweenness[w]+delta[w] 
					 
	# normalize 
	if normalize: 
		order=len(betweenness) 
		if order <=2: 
			return betweenness # no normalization b=0 for all nodes 
		scale=1.0/((order-1)*(order-2)) 
		for v in betweenness: 
			betweenness[v] *= scale 
 
   	return betweenness
					

def betweenness(G,normalize=True,edge_weight_fxn=None):
	"""
	Compute the betweenness centrality for nodes in G: 
	the fraction of number of shortests paths that pass through each node. 
 
	The keyword normalize (default=True) specifies whether the  
	betweenness values are normalized by b=b/(n-1)(n-2) where 
	n is the number of nodes in G. 
 
	The keyword edge_weight_fxn specifies a function that returns the weight
	of the edges.  By default, the function assumes unit weight.
	
	This function calls brandes_betweenness(...), which is an implementation
	of the betweenness centrality measure.
	"""
	return brandes_betweenness(G,normalize,edge_weight_fxn)

#####
# Pairwise disconnectivity functions
def __dfs_count_pairs(G):
	count = 0
	
	for s in G.V:
		visited = set()
		visited.add(s)
		
		queue = list()
		queue.append((s,G.neighbors_iter(s)))
		
		while len(queue) > 0:
			n,niter = queue[-1]
			
			# try to get a new neighbor
			v = None
			try:
				v = niter.next()
			except StopIteration: # no more neighbors!
				queue.pop()
			
			if v is not None and v not in visited:
				count += 1
				visited.add(v)
				queue.append((v,G.neighbors_iter(v)))

	if not G.is_directed():
		return count/2
	else:
		return count
	

def pairwise_disconnectivity(G,v):
	"""
	Compute the pairwise disconnectivity index for node v in G.
	
	If v is a list or set, then the pairwise disconnectivity index is computed
	for the set of vertices.
	
	This measure was first defined in paper:
	A. P. Potapov, B. Goemann, and E. Wingender.
	The pairwise disconnectivity index as a new metric for the topological
	analysis of regulatory networks.
	BMC Bioinformatics, 9:227, 2008.
	http://www.biomedcentral.com/1471-2105/9/227
	"""
	#####
	# compute N0
	N0 = float(__dfs_count_pairs(G))
		
	#####
	# Make the modified graph (missing v)
	G2 = None
	if G.is_directed():
		G2 = DiGraph(graph=G)
	else:
		G2 = Graph(graph=G)
	
	if type(v) == types.ListType:
		for n in v:
			G2.N.remove(n)
	else:
		G2.N.remove(v)
	
	#####
	# Compute N_v
	N_v = float(__dfs_count_pairs(G2))
	
	#####
	# Return disconnectivity
	return (N0 - N_v) / N0
	
def main():
	pass

if __name__ == '__main__':
	main()

