"""
This module provides clustering functions related to trees.
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

import types
import array
import math
from pyweaver.tree import *

__all__ = ['hcluster','SINGLE','COMPLETE']

########################################################
#
# Hierarchical Clustering
#
SINGLE = 'single'
COMPLETE = 'complete'

def compute_distance(c1,c2,D,dist_measure):
	"""
	Compute the distance between clusters c1 and c2
	"""
	global SINGLE
	global COMPLETE
	
	X = c1[2]
	Y = c2[2]
	
	if dist_measure == SINGLE:
		return min([D[x][y] for x in X for y in Y])
	elif dist_measure == COMPLETE:
		return max([D[x][y] for x in X for y in Y])
	else:
		raise Exception, 'Unknown distance measure %s' % dist_measure

def build_subtree(tree,parent,distance,cluster,labels):
	
	
	if type(cluster) == types.IntType:
		node = None
		
		if labels is not None:
			node = Node(item_idx = cluster, label = labels[cluster])
		else:
			node = Node(item_idx = cluster)

		tree.add_child(parent,node,distance)
	else: # len(cluster) == 3
		next_distance = cluster[0]
		node = Node(item_idx = None, label = '')
		
		tree.add_child(parent,node,distance)
		
		left = cluster[1]
		right = cluster[2]
		
		build_subtree(tree,node,next_distance,left,labels)
		build_subtree(tree,node,next_distance,right,labels)
		

def hcluster(D,distance_fxn=SINGLE,labels=None):
	"""
	Hierarchically cluster data with distances D, where D[i][j] is the distance between item i
	and item j.  In the cluster tree returned, leaf nodes will have a field 'item_idx' which gives
	the item index associated with that node.
	
	The distance_fxn is the method used to measure distances between clusters (since only the
	distances between the items themselves are given).  Valid values for this method are:
	
		- SINGLE - where the distance between X and Y is the minimum distance between any element in
		X and any element in Y.
		
		- COMPLETE - where the distance between X and Y is the maximum distance between any element
		in X and any element in Y.
	
	If labels is specified, then labels[i] is a name for the item with index i.  If labels is
	specified, then the leaf nodes in the cluster tree will also have a field 'label' which is
	the label of that item.
	
	The tree returned will be indexed by item_idx (and label if used) with item_idx being the
	primary index.
	"""
	
	#####
	# Setup...
	ngroups = len(D)
	
	if ngroups == 1:
		raise Exception, 'Only one item.  No clustering to do.'
	
	# make an array for caching cluster distances
	dist_lookup = [array.array('f',[-1.] * (2 * ngroups)) for i in range(0,2*ngroups)]
	
	#####
	# create the initial clusters
	# each cluster consists of a tuple: (cluster number, the hierarchical cluster, cluster members)
	clusters = [(x,x,[x]) for x in range(0,ngroups)]
	cnum = ngroups
	
	#####
	# coalesce the clusters until only one remains
	while len(clusters) > 1:
		min_pair = (0,1)
		min_dist = dist_lookup[clusters[0][0]][clusters[1][0]]
		
		if min_dist < 0:
			min_dist = compute_distance(clusters[0],clusters[1],D,distance_fxn)
			dist_lookup[clusters[0][0]][clusters[1][0]] = min_dist
		
		for i in range(0,len(clusters)):
			for j in range(i+1,len(clusters)):
				tmp_dist = dist_lookup[clusters[i][0]][clusters[j][0]]
				
				if tmp_dist < 0:
					tmp_dist = compute_distance(clusters[i],clusters[j],D,distance_fxn)
					dist_lookup[clusters[i][0]][clusters[j][0]] = tmp_dist
				
				if tmp_dist < min_dist:
					min_dist = tmp_dist
					min_pair = (i,j)
					
		# c2 must be popped before c1 or else the ordering changes.
		c2 = clusters.pop(min_pair[1])
		c1 = clusters.pop(min_pair[0])

		clusters.append((cnum,(float(min_dist)/2.,c1[1],c2[1]),c1[2] + c2[2]))
		cnum += 1
	
	#####
	# make the tree
	root = Node(item_idx = None, label = '')
	tree = RTree(root)
	if labels is None:
		tree.N.index('item_idx')
	else:
		tree.N.index('item_idx','label')
	root_cluster = clusters[0][1]
	distance = root_cluster[0]
	
	# build left subtree
	build_subtree(tree,root,distance,root_cluster[1],labels)
	# build right subtree
	build_subtree(tree,root,distance,root_cluster[2],labels)
	
	return tree