"""
This module provides utilities for inspecting the connectivity of a graph.
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

import array
from pyweaver.constants import BOTH, IN, OUT
import pyweaver.shortestpath as sp

__all__ = ['BOTH','IN','OUT','lcd','cd']

def lcd(G,directed=False,normalize=True):
	"""
	Compute the link connectivity distribution, Q.  Q[i][j] is the number of links that connect a node
	with degree i to a node with degree j.  If directed is True, then the link's source node
	must have degree i and the target node degree j.
	
	If normalize is True, then Q[i][j] is the *fraction* of the links in G that connect a node with
	degree i to a node with degree j.
	"""
	max_degree = G.degree()
	
	Q = [array.array('f',[0.] * (max_degree+1)) for i in range(max_degree+1)]
	
	for s in G.N:
		ks = G.degree(s)
		for e in G.E(s):
			t = G.target(e)
			kt = G.degree(t)
			
			Q[ks][kt] += 1
			
			if not directed:
				Q[kt][ks] += 1
	
	# normalize Q
	if normalize:
		num_links = float(len(G.E))
	
		for L in Q:
			for i in range(len(L)):
				L[i] = float(L[i]) / num_links
	
	return Q

def cd(G,directed=BOTH,normalize=True):
	"""
	Compute the connectivity distribution, D, for graph G where D[i] is 
	the number of nodes with i neighbors.
	
	If directed is BOTH, then both in and out neighbors are counted.
	If directed is IN, then only inbound edges are counted.
	If directed is OUT, then only outbound edges are counted.
	
	Setting directed to IN or OUT is only allowed when G is a directed graph.
	"""
	# sanity check
	if not G.is_directed() and (directed == IN or directed == OUT):
		raise Exception, 'Since G is undirected, argument "directed" must be BOTH, was %s' % str(directed)
	
	result = array.array('f')
	
	for n in G.N:
		num_neighbors = None
		if directed == BOTH:
			num_neighbors = G.degree(n)
		elif directed == IN:
			num_neighbors = G.indegree(n)
		elif directed == OUT:
			num_neighbors = G.outdegree(n)

		if num_neighbors >= len(result):
			result.extend([0.] * (num_neighbors - len(result) + 1))
		result[num_neighbors] += 1
	
	# normalize result if requested
	if normalize:
		num_edges = float(len(G.E))
		
		for i in range(len(result)):
			result[i] = result[i] / num_edges
	
	# done!
	return result

def efficiency(G,directed=False):
	"""
	Return the efficiency of this graph.  Graph efficiency, also known as the
	harmonic mean, is one measure of the average shortest paths present in the
	network.

	The efficiency of the directed or undirected versions of this graph can be
	computed by setting the directed parameter to the appropriate value.  In order
	to compute the directed efficiency, G must be directed.
	"""
	graph = G
	if directed:
		if not G.is_directed():
			raise Exception, 'Cannot compute directed efficiency on undirected graph'
			
		graph = Graph(graph=G)

	nnodes = float(len(graph.N))

	if nnodes == 0:
		return 0

	# compute sum of distances
	D = 0.
	node_list = graph.N()

	for i in range(0,len(node_list)):
		Dmap = sp.dijkstra_weight(graph,source = node_list[i])

		for j in range(i+1,len(node_list)):
			dist = Dmap[node_list[j]]

			if dist > 0:
				D += (1/float(dist))

	eff = (1 / (0.5 * nnodes * (nnodes+1))) * D

	return eff
		
#####
# For testing purposes...
def main():
	pass

if __name__ == '__main__':
	main()

