"""
This module contains a variety of functions for generating random graphs.
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

import types
import random as rnd

from pyweaver.constants import *
import pyweaver.graph as g_mod

__all__ = ['random', 'erdos_renyi', 'barabasi_albert']

def erdos_renyi(graph=UNDIRECTED,nodes=100,p=0.1,self_loops=False,edge_fxn=None):
	"""
	Generate an Erdos-Renyi network with a specifed number of nodes.  The parameter p specifies the probability
	of an edge being included in the graph.
	
	graph is either an empty undirected graph or UNDIRECTED.  If a graph is specified, then this graph will be
	populated using the Erdos-Renyi algorithm.
	
	If nodes is a number, then this number of nodes will be added to the network.  If nodes is a list, then
	the objects in the list will be used as nodes.
	
	If self_loops is True, then self-loops will be considered when adding edges.
	
	If edge_fxn is specified, then this function is used to generate the edges added to the graph.
	
	The graph constructed is returned.
	"""
	# get the graph
	if graph == UNDIRECTED:
		graph = g_mod.Graph()
	else:
		if len(graph.N) > 0 or len(graph.E) > 0:
			raise Exception, 'Graph must be empty'
	
	# get the nodes
	node_list = []
	if type(nodes) == types.ListType:
		node_list = list(nodes)
	else:
		node_list = range(nodes)
		
	# add the nodes to the graph
	graph.V.update(node_list)
		
	# add edges
	for i in range(len(node_list)):
		ni = node_list[i]
		
		first_j = i
		if not self_loops:
			first_j += 1
			
		for j in range(first_j,len(node_list)):
			
			if rnd.random() <= p:
				nj = node_list[j]
				
				if edge_fxn is not None:
					graph.E.add(ni,nj,edge_fxn())
				else:
					graph.E.add(ni,nj)
	
	return graph

def barabasi_albert(graph=UNDIRECTED,nodes=100,m=2):
	"""
	Use the Barabasi-Albert method for constructing a scale-free graph.  If an
	undirected graph is specified, then the method will extend the existing topology,
	adding the specified number of nodes.
	
	If the graph is not specified, then the method will construct an undirected graph having
	two nodes.  This network will be extended until the number of nodes specified is 
	reached.
	
	When a node is added to the graph, it will be connected to m other nodes, selected by
	favoring those with higher degree.
	
	If nodes is a number, it specifies the number of new nodes to add to the graph (these
	nodes will each be an integer between 0 and nodes).  If nodes is a list, then the objects
	in the list are used as the nodes.
	"""
	
	#####
	# get the node list to add
	node_list = []
	
	if type(nodes) == types.ListType:
		node_list = list(nodes)
	else:
		node_list = range(nodes)
		
	
	#####
	# Build the graph if necessary
	if graph == UNDIRECTED:
		graph = g_mod.Graph()
		
		# add 2 nodes
		n1 = node_list.pop()
		n2 = node_list.pop()
		graph.E.add(n1,n2)
	elif graph.is_directed():
		raise Exception, 'Barabasi-Albert method is only defined over undirected graphs'
	
	total_degree = 2
		
	#####
	# add nodes
	nodes = graph.V()
	
	while len(node_list) > 0:
		n = node_list.pop()
		
		graph.V.add(n)
		
		# attach the node
		for i in range(m):
			
			# pick a node
			other_node = None
			x = rnd.random()
			prob = 0.0
			
			rnd.shuffle(nodes)
			while other_node == None:
				for n2 in nodes:
					prob += float(graph.degree(n2)) / float(total_degree)
				
					if x < prob and not graph.E.contains(n,n2):
						other_node = n2
						break
			
			graph.E.add(n,other_node)
			total_degree += 2
	
		nodes.append(n)
	
	# done!
	return graph

def randomize_connectivity(source,dest=None,self_loops=True,degree=True,io_degree=False):
	"""
	Generate a randomized version of the source graph in which the only property that has changed
	is the connectivity.  All nodes and edges remain.  The only change is in which nodes
	any given edge connects.  dest, if specified, provides a graph that the randomized topology will 
	be placed in.  How edges are used to 'randomly' connect nodes depends on the properties
	that must be maintained:
	
		- Self loops (edges from a node to itself) are only allowed if self_loops is True (by default, self loops
	  	are allowed).
		
		- Degree maintains the degree of each node under the new connectivity.
		
		- I/O degree (io_degree) maintains the overall in degree and out degree of each node under the
		new connectivity.  Since in and out degrees are only applicable for directed networks, the
		values of io_degree is ignored for undirected networks, though a warning is displayed if this
		parameter is set.  If io_degree is True, then degree must be True.  If this condition is not met,
		an exception is thrown.
		
	This function returns the newly generated network.
	"""
	
	# some checks...
	if source.is_directed() == False and io_degree == True:
		raise Exception, 'io_degree cannot be True for an undirected source graph.'
	
	if io_degree and not degree:
		raise Exception, 'If io_degree is True, then degree must also be True (%s)' % str(source)
	
	#####
	# setup the destination graph
	if dest is None:
		if source.is_directed():
			dest = g_mod.DiGraph()
		else:
			dest = g_mod.Graph()
	
	if source.is_directed() != dest.is_directed():
		raise Exception, 'Source and destination graphs must both have the same directed-ness'
	
	nodes = source.N()
	edges = source.E()
	
	# add all nodes
	for n in nodes:
		dest.N.add(n)
	
	#####
	# create connectivity
	directed = dest.is_directed()
	if not degree:
		#####
		# Generate graph irrespective of original degree of nodes.
		for e in edges:
			n1 = nodes[random.randrange(len(free_nodes))]
			n2 = nodes[random.randrange(len(free_nodes))]

			while n1 == n2 and not self_loops:
				n1 = nodes[random.randrange(len(free_nodes))]
				n2 = nodes[random.randrange(len(free_nodes))]

			dest.E.add(n1,n2,e)
			
		return dest
	elif not io_degree:
		#####
		# Respect node connectivity, but not in/out degrees
		free_nodes = [(n,source.degree(n)) for n in nodes if source.degree(n) > 0]
		
		for e in edges:
			n1,ne1 = free_nodes[random.randrange(len(free_nodes))]
			n2,ne2 = free_nodes[random.randrange(len(free_nodes))]

			while n1 == n2 and (not self_loops or ne1 == 1):
				n1,ne1 = free_nodes[random.randrange(len(free_nodes))]
				n2,ne2 = free_nodes[random.randrange(len(free_nodes))]
	
			dest.E.add(n1,n2,e)
			
			if dest.degree(n1) == ne1:
				free_nodes.remove((n1,ne1))
			if dest.degree(n2) == ne2:
				free_nodes.remove((n2,ne2))
				
		return dest
	else:
		#####
		# Respect in/out degrees of nodes
		free_in_nodes = [(n,source.indegree(n)) for n in nodes if source.indegree(n) > 0]
		free_out_nodes = [(n,source.outdegree(n)) for n in nodes if source.outdegree(n) > 0]
		
		for e in edges:
			n1,ne1 = free_out_nodes[random.randrange(len(free_out_nodes))]
			n2,ne2 = free_in_nodes[random.randrange(len(free_in_nodes))]

			while n1 == n2 and not self_loops:
				n1,ne1 = free_out_nodes[random.randrange(len(free_out_nodes))]
				n2,ne2 = free_in_nodes[random.randrange(len(free_in_nodes))]
	
			dest.E.add(n1,n2,e)
			
			if dest.outdegree(n1) == ne1:
				free_out_nodes.remove((n1,ne1))
			if dest.indegree(n2) == ne2:
				free_in_nodes.remove((n2,ne2))
				
		return dest
			

def random(graph=UNDIRECTED,nodes=100,edges=100,self_loops=False):
	"""
	Generate a random graph.

	graph can be either DIRECTED, UNDIRECTED, or an actual graph object.  By default, it is
	UNDIRECTED.

	The value of directed determines how edges are constructed.  If directed is None, then
	the the type of edges constructed defers to the type of graph.  If directed is True, then
	(x,y) and (y,x) can both be added to the same graph.  If directed is False, then (x,y)
	and (y,x) are the same edge.  An exception will be thrown if the graph type and the
	value of directed conflict (if graph is undirected and directed = True).

	If nodes is a number, then that number of nodes will be used.  If nodes is a collection
	of objects, then those objects will be used as the nodes.
	
	If edges is a number, then that number of edges will be used.  If edges is a collection
	of objects, then those objects will be the decorators for the edges.
	
	If self_loops is True, then self loops are allowed.  Otherwise, self loops are not
	permitted.
		
	The randomly generated graph is returned.
	"""
	
	# Create the graph if necessary
	if graph == UNDIRECTED:
		graph = g_mod.Graph()
	elif graph == DIRECTED:
		graph = g_mod.DiGraph()

	# if directed is default, set it
	directed = graph.is_directed()
	
	####
	# generate the graph
	
	# build node list if necessary
	node_list = None
	num_nodes = 0
	if type(nodes) == types.IntType:
		node_list = range(0,nodes)
		num_nodes = nodes
	else:
		node_list = nodes
		num_nodes = len(nodes)
	
	# build edge decorator list if necessary
	ed_list = None
	if type(edges) == types.IntType:
		ed_list = range(0,edges)
	else:
		ed_list = edges

	# add all nodes in the node list
	for n in node_list:
		if n not in graph.N:
			graph.N.add(n)

	# generate the edges
	for ed in ed_list:		
		success = False
		
		while not success:
			# get two random nodes to connect
			u = node_list[rnd.randint(0,num_nodes-1)]
			v = node_list[rnd.randint(0,num_nodes-1)]
	
			# enforce self loop policy
			if u == v and not self_loops:
				while v == u:
					v = node_list[rnd.randint(0,num_nodes-1)]
			
			# check whether this edge is allowed
			if (directed and graph.E.contains(u,v)) or (not directed and (graph.E.contains(u,v) or graph.E.contains(v,u))):
				pass
			else:
				graph.E.add(u,v,ed)
				success = True
	
	# done! Return the graph
	return graph
