"""
This module provides support for reading graphs from and writing graphs to the DOT format.

For details on the DOT format, visit the url: http://www.graphviz.org/doc/info/lang.html
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

import types
import re

from pyweaver.constants import ALL
from pyweaver.graph import DiGraph, Graph, Node, Edge

__all__ = ['write','read', 'read_all', 'read_iter']

def read_iter(fhandle,generate_node=lambda x: Node(x),generate_edge=lambda: Edge(),generate_graph=None):
	"""
	Returns an iterator that reads a series of DOT-formatted networks from the file-like object 
	specified.  It is assumed that this file-like object contains *only* graphs in DOT format.  
	Graphs are read sequentially until the end of file is reached.
	"""
	while True:
		Gin = None
		if generate_graph is not None:
			Gin = generate_graph()
	
		G = read_graph(fhandle, generate_node, generate_edge, Gin, False)
	
		if G is None:
			break
		else:
			yield G

def read_all(fhandle,generate_node=lambda x: Node(x),generate_edge=lambda: Edge(),generate_graph=None):
	"""
	Read graphs from a file-like object.  It is assumed that this file-like object contains *only*
	graphs in DOT format.  Graphs are read sequentially until the end of file is reached.  A list
	of the graphs is returned.  The graphs are stored in the order in which they were read.
	
	For documentation of function arguments, see the read(...) function, since this is used
	internally.
	"""
	
	graph_list = []
	
	while True:
		Gin = None
		if generate_graph is not None:
			Gin = generate_graph()
	
		G = read_graph(fhandle, generate_node, generate_edge, Gin, False)
	
		if G is None:
			break
		else:
			graph_list.append(G)
			
	return graph_list

def read(fhandle,generate_node=lambda x: Node(x),generate_edge=lambda: Edge(), G=None):
	"""
	Read a graph from a file-like object.  The node/edge objects added to the new graph
	are constructed by the node and edge factory functions generate_node and generate_edge
	generate_node function takes one argument which is the label for the node.  The edge
	function takes zero arguments.  The attributes specified in the DOT file will be 
	added to these objects.
	
	An existing graph object, G, can be specified.  The graph read will be loaded into this
	object.  The directedness of the graph object must agree with the directedness of
	the graph data being read.
	
	This reader currently does not support the following aspects of the DOT format:
	  - subgraphs
	  - mixed graphs
	  - comments
	  - HTML strings
	  - ports
	"""
	
	G = read_graph(fhandle, generate_node, generate_edge, G)
	
	return G

def write(G,fhandle,graph_attributes=None,node_attributes=None,edge_attributes=None):
	"""
	Write the graph G to the file-like object fhandle using the DOT format.  By default
	no attributes of the graph, edges, or nodes are written out.  The set of attributes can
	be specified by setting the graph/node/edge_attributes to be a list of the names of
	attributes that should be included in the output.  pyweaver.constants.ALL can also be
	used to indicate that all attributes should be written.
	"""

	directed = G.is_directed()
	
	# start the graph block
	if directed:
		fhandle.write('strict digraph {\n')
	else:
		fhandle.write('strict graph {\n')
	
	# write graph attributes
	if graph_attributes is not None and hasattr(G,'__dict__'):
		attrs = None
		
		if graph_attributes == ALL:
			attrs = [(k,v) for k,v in G.__dict__.items() if type(v) is not types.FunctionType]
		else:
			attrs = [(k,v) for k,v in G.__dict__.items() if type(v) is not types.FunctionType and k in graph_attributes]
		
		for name,value in attrs:
			fhandle.write('\t' + name + ' = ' + pyvalue2DOTstr(value) + ';\n')
			
	# a little whitespace for readability
	fhandle.write('\n')
	
	# write node attributes
	if node_attributes is not None:
		for n in G.N:
			attrstr = node_edge2attrstr(n, node_attributes)
			
			if attrstr is None:
				continue
			
			fhandle.write('\t"' + str(n) + '" ' + attrstr + ';\n');

	# write edges
	for e in G.E:
		conn_str = None
		
		if directed:
			n1 = G.source(e)
			n2 = G.target(e)
			
			conn_str = '"' + str(n1) + '" -> "' + str(n2) + '"'
		else:
			n1,n2 = G.endpoints(e)
			conn_str = '"' + str(n1) + '" -- "' + str(n2) + '"'
			
		fhandle.write('\t' + conn_str)
			
		# write attributes
		if edge_attributes is not None:
			attrstr = node_edge2attrstr(e, edge_attributes)
			if attrstr is None:
				fhandle.write(';\n')
			else:
				fhandle.write(' ' + attrstr + ';\n')
		else:
			fhandle.write(';\n')
			
	# end the graph block
	fhandle.write('}')
	
	return
		
######
# DOT file writing helper functions
######

def node_edge2attrstr(obj,attr_list = ALL):
	"""
	Return the attribute block for a node or edge.  By default the block contains all dynamic properties of the
	object that are not functions.  The set of attributes can be specified using the attr_list - a list containing
	the names of the attributes to include.
	"""
	if not hasattr(obj,'__dict__'):
		return None
	
	attrs = None
	
	if attr_list == ALL:
		attrs = [(k,v) for k,v in obj.__dict__.items() if type(v) is not types.FunctionType]
	else:
		attrs = [(k,obj.__dict__[k]) for k in attr_list if k in obj.__dict__]
	
	if len(attrs) == 0:
		return None

	attr_str = '['
	
	first = True
	for name,value in attrs:
		if not first:
			attr_str += ', '
		else:
			first = False
			
		attr_str += name + ' = ' + pyvalue2DOTstr(value)
	
	attr_str += ']'
	
	return attr_str

def pyvalue2DOTstr(value):
	"""
	Convert a python value into a DOT value.
	"""
	if type(value) == types.IntType or type(value) == types.FloatType:
		return str(value)
	else:
		return '"' + str(value) + '"'

#########
# DOT file reading helper functions and constants
#########
	
SEMICOLON = ';'
LBRACKET = '{'
RBRACKET = '}'
LBRACE = '['
RBRACE = ']'
EQUALS = '='
WORD = 'word'
NUMBER = 'num'
STRING = '"'
GRAPH = 'graph'
STRICT = 'strict'
DIGRAPH = 'digraph'
SUBGRAPH = 'subgraph'
NODE = 'node'
EDGE = 'edge'
COMMA = ','
UEDGE = '--'
DEDGE = '->'
	
def read_graph(fhandle,generate_node,generate_edge,G=None,error_on_nograph=True):
	"""
	The top-level method that coordinates and organizes the parsing of DOT files.
	"""
	# read tokens until a matching top-level curly-brace is found
	token_list = cache_tokens(fhandle)
	
	# sanity check - look for a last bracket
	if len(token_list) == 0:
		if error_on_nograph:
			raise Exception, 'No tokens to read.'
		else:
			return None
		
	if token_list[-1][0] != RBRACKET:
		raise Exception, 'Invalid DOT format.  The graph block is not terminated.'
		
	# get started!
	tidx = 0
	directed = False

	# we don't really care about the strict keyword, since we assume a strict
	# interpretation of the directedness of graphs anyway
	if token_list[tidx][0] == STRICT:
		tidx += 1
	
	# parse directedness
	tok = token_list[tidx]
	tidx += 1
	
	if tok[0] == DIGRAPH:
		directed = True
	elif tok[0] == GRAPH:
		directed = False
	else:
		raise Exception, 'Invalid DOT format.  Expected GRAPH or DIGRAPH, found %s' % tok[1]

	# make a graph if necessary
	if G is not None:
		if G.is_directed() != directed:
			raise Exception, 'Graph supplied does not agree with directedness of graph being read'
	else:
		if directed:
			G = DiGraph()
		else:
			G = Graph()
	
	# set the label of the graph if it is given
	if is_ID(token_list[tidx]):
		G.label = token_list[tidx][1]
		tidx += 1
	
	# parse a left bracket
	tok = token_list[tidx]
	tidx += 1
	
	if tok[0] != LBRACKET:
		raise Exception, 'Invalid DOT format.  Expected "{", found ' + tok[1]
	
	# start reading statements
	node_attrs = {}
	edge_attrs = {}
	while token_list[tidx][0] != RBRACKET:
		tidx = process_stmt(token_list, tidx, G, generate_node, generate_edge, node_attrs, edge_attrs)

	return G

def process_stmt(token_list, tidx, G, generate_node, generate_edge, node_attrs, edge_attrs):
	"""
	This function parses and stores the content of a single statement into the graph provided.  It returns the new token index.
	"""
	# handle statements by type
	ttype = token_list[tidx][0]
	
	if is_ID(token_list[tidx]):
		ttype = token_list[tidx+1][0]
		
		if ttype == EQUALS:
			tidx = process_attr_stmt(token_list, tidx, G, node_attrs, edge_attrs)
		elif ttype == LBRACE:
			tidx = process_node_stmt(token_list, tidx, G, generate_node, node_attrs)
		elif ttype == UEDGE or ttype == DEDGE:
			tidx = process_edge_stmt(token_list, tidx, G, generate_node, node_attrs, generate_edge, edge_attrs)
	elif ttype == NODE:
		tidx = process_attr_stmt(token_list, tidx, G, node_attrs, edge_attrs)
	elif ttype == EDGE:
		tidx = process_attr_stmt(token_list, tidx, G, node_attrs, edge_attrs)
	elif ttype == GRAPH:
		tidx = process_attr_stmt(token_list, tidx, G, node_attrs, edge_attrs)
	elif ttype == SUBGRAPH:
		raise Exception, 'Subgraph statements not currently supported'
	else:
		raise Exception, 'Illegal statement opening: ' + token_list[tidx][1]
	
	if token_list[tidx][0] != SEMICOLON:
		raise Exception, 'Expected ";", found ' + token_list[tidx][1]
	
	tidx += 1
	return tidx

def process_edge_stmt(token_list, tidx, G, generate_node, node_attrs, generate_edge, edge_attrs):
	"""
	This function parses the edge declaration statement.  It returns the new token index.
	"""
	# edge type sanity check
	if (token_list[tidx+1][0] == UEDGE and G.is_directed()) or (token_list[tidx+1][0] == DEDGE and not G.is_directed()):
		raise Exception, 'Edge type is incompatible with graph type: ' + token_list[tidx+1][0]

	if not is_ID(token_list[tidx+2]):
		raise Exception, 'Edge statement requires two node IDs.  Expected a node ID, found ' + token_list[tidx+2][1]
	
	# create the edge
	edge = generate_edge()
	n1_label = token_list[tidx][1]
	n1_list = [n for n in G.N if n.label == n1_label]
	n1 = None
	
	if len(n1_list) == 0:
		n1 = generate_node(token_list[tidx][1])
		
		# add all default node attributes
		for k,v in node_attrs.items():
			n1.__dict__[k] = v
	else:
		n1 = n1_list[0]
		
	n2_label = token_list[tidx+2][1]
	n2_list = [n for n in G.N if n.label == n2_label]
	n2 = None
	
	if len(n2_list) == 0:
		n2 = generate_node(token_list[tidx+2][1])
		
		# add all default node attributes
		for k,v in node_attrs.items():
			n2.__dict__[k] = v
	else:
		n2 = n2_list[0]
	
	tidx += 3
	
	# add all default edge attributes
	for k,v in edge_attrs.items():
		edge.__dict__[k] = v
	
	# parse edge attributes
	if token_list[tidx][0] == LBRACE:
		tidx,attrs = parse_attr_list(token_list, tidx)
		
		for k, v in attrs.items():
			edge.__dict__[k] = v
				
	G.E.add(n1,n2,edge)			
			
	return tidx

def process_node_stmt(token_list, tidx, G, generate_node, node_attrs):
	"""
	Process a statement that defines a node.
	"""
	node = generate_node(token_list[tidx][1])
	tidx += 1
	
	for k, v in node_attrs.items():
		node.__dict__[k] = v
		
	
	if token_list[tidx][0] == LBRACE:
		tidx,attrs = parse_attr_list(token_list, tidx)
		
		for k, v in attrs.items():
			node.__dict__[k] = v
	
	G.N.add(node)
			
	return tidx

def process_attr_stmt(token_list, tidx, G, node_attrs, edge_attrs):
	"""
	Process a statement that defines a default attribute for a graph, node, or edge.
	"""
	ttype = token_list[tidx][0]
	
	if is_ID(token_list[tidx]) and token_list[tidx+1][0] == EQUALS:
		G.__dict__[token_list[tidx][1]] = token_list[tidx+2][1]
		return tidx+3
	elif ttype == NODE:
		tidx,attrs = parse_attr_list(token_list, tidx + 1)
		for k, v in attrs.items():
			node_attrs[k] = v
	elif ttype == EDGE:
		tidx,attrs = parse_attr_list(token_list, tidx + 1)
		for k, v in attrs.items():
			edge_attrs[k] = v		
	elif ttype == GRAPH:
		tidx,attrs = parse_attr_list(token_list, tidx + 1)
		for k, v in attrs.items():
			G.__dict__[k] = v

	return tidx

def parse_attr_list(token_list, tidx):
	"""
	Parse an attribute block and return the mapping in the form of a dictionary.
	"""
	attrs = {}
	
	while token_list[tidx][0] == LBRACE or token_list[tidx][0] == COMMA:
		tidx += 1
		
		if not is_ID(token_list[tidx]):
			raise Exception, 'Expected an identifier, found ' + token_list[tidx][1]
		key = token_list[tidx][1]
		
		tidx += 1

		if not token_list[tidx][0] == EQUALS:
			raise Exception, 'Expected "=", found ' + token_list[tidx][1]
		
		tidx += 1

		value = token_list[tidx][1]
		attrs[key] = value
		
		tidx += 1
		
	if token_list[tidx][0] != RBRACE:
		raise Exception, 'Expected "]", found ' + token_list[tidx][1]
	
	tidx += 1
		
	return tidx,attrs
		
def is_ID(token):
	"""
	Return true if a token is one of the types that an ID can be
	"""
	return (token[0] == WORD or token[0] == NUMBER or token[0] == STRING)

def cache_tokens(fhandle):
	"""
	Create a token list that ends with either the last token in the file or the first root-level
	matching curly braces, whichever comes first.
	"""
	
	num_pattern = re.compile(r'-?(\.[0-9]+|[0-9]+(\.[0-9]+)?)')
	
	token_list = []
	depth = 0
	seen_bracket = False
	
	next_char = None
	while True:
		
		if next_char:
			c,next_char = next_char,None
		else:
			c = fhandle.read(1)
	
		if len(c) == 0:
			break
		if c.isspace():
			continue
		elif c == SEMICOLON or c == LBRACKET or c == RBRACKET or c == LBRACE or c == RBRACE or c == EQUALS or c == COMMA:
			token_list.append((c,c))
		elif c == STRING:

			# read to the closing quote
			quote = ''
			c = fhandle.read(1)
			while not c == '"':
				quote += c
				c = fhandle.read(1)
				
			token_list.append((STRING,quote))
		elif c.isdigit() or c == '-' or c == '.':

			# read to the next whitespace
			num_str = c
			c = fhandle.read(1)
			
			if num_str == '-' and c == '-':
				token_list.append((UEDGE,'--'))
			elif num_str == '-' and c == '>':
				token_list.append((DEDGE,'->'))
			else:
				while not c.isspace() and not is_reserved_char(c):
					num_str += c
					c = fhandle.read(1)
					
				if is_reserved_char(c):
					next_char = c
					
				if re.match(num_pattern, num_str) is None:
					raise Exception, 'Unparsable number-like string: ' + num_str
				
				num = None
				if num_str.find('.') == -1:
					num = int(num_str)
				else:
					num = float(num_str)
				
				token_list.append((NUMBER,num))
		elif c.isalpha():
			
			# read to the next whitespace
			word = c
			c = fhandle.read(1)
			while not c.isspace() and not is_reserved_char(c):
				word += c
				c = fhandle.read(1)
			
			if is_reserved_char(c):
				next_char = c
				
			if word.lower() == 'graph':
				token_list.append((GRAPH,GRAPH))
			elif word.lower() == 'strict':
				token_list.append((STRICT,STRICT))
			elif word.lower() == 'digraph':
				token_list.append((DIGRAPH,DIGRAPH))				
			elif word.lower() == 'node':
				token_list.append((NODE,NODE))		
			elif word.lower() == 'edge':
				token_list.append((EDGE,EDGE)) 
			elif word.lower() == 'subgraph':
				token_list.append((SUBGRAPH,SUBGRAPH))
			else:
				token_list.append((WORD,word))

		# update bracket depth
		if token_list[-1][0] == LBRACKET:
			depth += 1
			seen_bracket = True
		elif token_list[-1][0] == RBRACKET:
			depth -= 1
				
		# check for the stopping condition
		if seen_bracket and depth == 0:
			break
	
	return token_list

def is_reserved_char(c):
	"""
	Return true if the character is a reserved character.
	"""
	return (c == SEMICOLON or c == LBRACKET or c == RBRACKET or c == LBRACE or c == RBRACE or c == COMMA or c == EQUALS or c == '-')