"""
This module provides support for reading from and writing to edge lists.  An edge list format is a list of
edges distinguished by their endpoints.  Each line of the edgelist specifies exactly one edge.  A line
can have two formats:

	eid nid1 nid2
	
	nid1 nid2
	
In the first case, an edge label is given.  In the second case, only the node labels are given.
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""
from pyweaver.constants import *
from pyweaver.graph import Graph, DiGraph, Node, Edge

__all__ = ['write','read']

def read(fh, G = UNDIRECTED, node_fxn = lambda x: Node(x), edge_fxn = lambda x: Edge(label=x), num_edges = None):
	"""
	Read graph data from the file-like object in the edge list format.  Nodes are constructed
	using the function node_fxn which takes the node label.  If the node does not have a label
	then the label is None.  The same rules apply to the edge function edge_fxn.
	
	If G is the UNDIRECTED constant, then an undirected graph is read.  If G is the DIRECTED
	constant then the edges are read as directed edges.  If G is a graph object, then the
	edges are read according to the directedness of the graph object.
	
	If num_edges is a positive integer, then this number of edges are read from the file-like
	object before the function returns.  If None, this function reads edges until it reaches
	the end of the character stream.
	"""
	# initialize the graph object if necessary
	if G == UNDIRECTED:
		G = Graph()
	elif G == DIRECTED:
		G = DiGraph()
	
	# read each line
	label2node = {}
	
	def label_to_node(label):
		if label in label2node:
			return label2node[label]
		else:
			n = node_fxn(label)
			G.V.add(n)
			label2node[label] = n
			return n
	
	if num_edges == 0:
		return G
	
	for line in fh:
		data = line.strip().split()
		
		if len(data) == 2:
			n1 = label_to_node(data[0])
			n2 = label_to_node(data[1])
			
			G.E.add(n1,n2)
		elif len(data) == 3:
			n1 = label_to_node(data[0])
			n2 = label_to_node(data[1])
			e = edge_fxn(data[2])
			
			G.E.add(n1,n2,e)
		else:
			raise Exception, 'Line "%s" is not violates the edge list format' % line.strip()
		
		# enforce number of edges to read
		if num_edges is not None:
			num_edges -= 1
			if num_edges == 0:
				break
	
	return G

def write(G, fh, node_label_fxn = lambda x: x.label, edge_label_fxn = None):
	"""
	Write the graph, G, to the file-like object in the edge list format such that the node
	label used for node n is node_label_fxn(n).  By default, this is the LABEL_FXN, evaluating
	to the value of n.label.  Other valid values can be custom functions that takes
	a node as input and returns a string as output.
	
	If edge_label_fxn is None, then no edge labels are written to the output.  Otherwise, 
	edge_label_fxn(e) is used as the edge label for edge e.  If this is None, then no edge
	label is written for this edge.
	"""
	
	for e in G.E:
		n1, n2 = G.endpoints(e)
		
		if edge_label_fxn is None:
			fh.write('%s %s\n' % (node_label_fxn(n1),node_label_fxn(n2)))
		else:
			elabel = edge_label_fxn(e)
			
			if elabel is None:
				fh.write('%s %s\n' % (node_label_fxn(n1),node_label_fxn(n2)))
			else:
				fh.write('%s %s %s\n' % (elabel, node_label_fxn(n1),node_label_fxn(n2)))
				
	# done!
	return
	
def main():
	pass


if __name__ == '__main__':
	main()

