"""
This module provides support for reading graphs from and writing graphs to the CSV (comma-separated value) file
format.  Graphs can be stored to a CSV in a variety of ways.  This module provides support for reading and writing
the following formats:

  - Connectivity matrix
  
Future versions will support loading edge lists.
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

from pyweaver.graph import Graph, DiGraph, Node, Edge

import csv

__all__ = ['read_cmatrix','write_cmatrix']

def generate_node(x):
	n = Node()
	n.label = x
	return n	

def read_cmatrix(fhandle,generate_node=generate_node,generate_edge=lambda: Edge(),directed=False,sep=','):
	"""
	Read a graph from the connectivity matrix stored in the file-like object fhandle.
	The graph constructed and read is either directed or undirected.  If directed, then
	edges are assumed to be directional and the values of c[x][y] and c[y][x] are processed
	as separate edges.  If undirected, then only the value of c[x][y] is used.  Weights will
	be parsed as floats.  Only missing values for c[x][y] are interpretted as there being no edge
	from x to y.

	The nodes and edges added to the graph are constructed using the node/edge factory
	functions generate_node and generate_edge.  generate_node is a function that takes the
	name of the node as its only argument.  generate_edge is a function that takes no
	arguments.
	
	The actual value of c[x][y] is the weight of the edge.  In the graph returned, node n, has
	a field n.label to designate the name of the node.  Edges have no identifiers, but do have
	a weight: e.weight.
	"""
	
	if directed:
		G = DiGraph()
	else:
		G = Graph()
		
	reader = csv.reader(fhandle,delimiter=sep)

	hrow = None
	nodes = {}
	
	for row in reader:
		if hrow == None:
			hrow = row
			
			for name in hrow[1:]:
				n = Node()
				n.label = name
				
				G.N.add(n)
				nodes[n.label] = n
			
			continue
		
		xlbl = row[0]
		x = nodes[xlbl]
		for w,i in zip(row[1:],range(1,len(row))):
			ylbl = hrow[i]
			y = nodes[ylbl]
			
			# skip all edges that have no weight
			if len(w) == 0 or w.isspace():
				continue
			
			e = Edge()
			e.weight = float(w)
			G.E.add(x,y,e)
			
	return G

def write_cmatrix(G,fhandle,edge_weight=lambda x: 1,node_name=lambda x: str(x), sep=',', noexist_value=0):
	"""
	Write the graph G to the file-like object fhandle as a connectivity
	matrix.  The value of entry i-j in the connectivity matrix is the weight of the
	edge connecting node i and j.  The edge weight is determined by the edge_weight
	function specified.  By default, the function used assigns 1 to all edges that
	exist.  The value used for a non-existent edge is given by noexist_value.
	In undirected graphs, the matrix is symmetric across the diagonal.  In directed
	graphs the matrix is not necessarily symmetric.  The first column and the first
	row of the CSV file provide the node names that are being connected.  By default
	node names are simply the str(...) representation of the node.  This can be
	modified by specifying the node_name function that takes a node and returns its 
	string name.
	
	Since the connectivity format can only represent one undirected/two directed edges
	per pair, any edges beyond this number will trigger an exception.
	"""
	
	# get a list of nodes and node names
	nlist = G.N()
	nnames = [node_name(n) for n in nlist]
	
	# write the names as headers
	fhandle.write(sep.join([''] + nnames) + '\n')
	
	# write the matrix
	for n,name in zip(nlist,nnames):
		row = [name]
		
		for n2 in nlist:
			es = G.E.get(n,n2)
			
			if len(es) > 1:
				raise Exception, 'More than one edge exists between %s and %s' % (node_name(n),node_name(n2))
			
			if len(es) == 1:
				row.append(str(edge_weight(es[0])))
			else:
				row.append(str(noexist_value))
		
		fhandle.write(sep.join(row) + '\n')
	
	# done
	return
