"""
This module provides support for reading biochemical networks from the KEGG Markup Language (KGML).

For details on the KGML format, visit the url: http://www.genome.jp/kegg/xml/
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

from pyweaver.graph import DiGraph, Node, Edge

from xml.dom import minidom

__all__ = ['read']

def read(fhandle,generate_node=lambda x: Node(x),generate_edge=lambda: Edge(),G=None):
	"""
	This function reads a KGML formatted network from the file-like object
	fhandle.  Nodes are generated via the generate_node function and edges
	by the generate_edge function.  Unless the graph parameter is specified,
	a new DiGraph is constructed.  All node, edge, and network properties are
	added as new fields to the node, edge, and graph respectively.
	
	This reader excludes compound relations as these can't be expressed correctly in
	a directed graph.  This issue will be addressed in a future version.
	"""
	kgmldoc = minidom.parse(fhandle)
	
	N = None
	if G is not None:
		N = G
	else:
		N = DiGraph()
	
	# get the first pathway element
	pathway_node = kgmldoc.getElementsByTagName('pathway')[0]
	
	#####
	# process all entries
	entry_nodes = pathway_node.getElementsByTagName('entry')
	
	id2node = {}

	for entry in entry_nodes:
		n = generate_node(entry.getAttribute('name'))
		n.id = entry.getAttribute('id')
		
		N.N.add(n)
		id2node[entry.getAttribute('id')] = n
	
	#####
	# process all relations
	relation_nodes = pathway_node.getElementsByTagName('relation')
	
	for r in relation_nodes:		
		rtype = None
		for c in r.getElementsByTagName('subtype'):
			name = c.getAttribute('name')
			if name == 'activation':
				rtype = 'activation'
				break
			elif name == 'inhibition':
				rtype = 'inhibition'
				break
		
		# skip unsupported types
		if rtype is None:
			continue
		
		n1 = id2node[r.getAttribute('entry1')]
		n2 = id2node[r.getAttribute('entry2')]

		e = generate_edge()
		e.type = rtype
		N.E.add(n1,n2,e)
		
	return N