"""
This module provides support for reading and writing trees in the Newick format.
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys
import os
import StringIO

from pyweaver.tree import *

__all__ = ['read_str','read']

NO_NAME = ''
NO_DISTANCE = None

def read_str(newick_str,node_generator=lambda x: Node(x),tree_generator=lambda x: RTree(x)):
	"""
	Read a tree in the network format from a string.
	"""
	return read(StringIO.StringIO(newick_str),node_generator,tree_generator)
	
def read(fh,node_generator=lambda x: Node(x),tree_generator=lambda x: RTree(x)):
	"""
	Read a tree in from the specified file-like object.
	
	Currently, this method only reads rooted trees.
	
	The type of node added to the tree can be specified by setting the node_generator,
	a function that takes one value - the label of the node.
	
	The type of tree that will be populated and returned can also be set by providing
	a tree_generator, a function that accepts the root of the tree and returns the
	a rooted-tree like object.
	"""
	
	token_list = make_token_list(fh)
	data_tree, tidx = read_node(token_list,0,node_generator)

	# store the data tree in the tree
	T = store_data(None,data_tree,tree_generator,None)
	
	return T

def store_data(parent,data,tree_fxn,T):
	
	node, distance, children = data
	
	if T is None:
		if distance != None:
			print 'WARNING: Distance was provided for the root, but was not used.'
		T = tree_fxn(node)
	else:
		T.add_child(parent, node, distance)
	
	for c in children:
		store_data(node, c, None, T)
		
	return T

def read_node(token_list,tidx,node_fxn):
	
	# is this node a leaf.  If not, read children...
	children = []
	if token_list[tidx][0] == LPAREN:
		while token_list[tidx][0] != RPAREN:
			
			if len(children) > 0 and token_list[tidx][0] != COMMA:
				raise Exception, 'Comma must separate siblings in a tree'
			tidx += 1
			
			child, tidx = read_node(token_list, tidx, node_fxn)
			children.append(child)
		tidx += 1
	
	# read the name
	name = NO_NAME
	if tidx < len(token_list) and token_list[tidx][0] == STRING:
		name = token_list[tidx][1]
		tidx += 1
	
	node = node_fxn(name)
	
	# read the distance
	distance = NO_DISTANCE
	if tidx < len(token_list) and token_list[tidx][0] == COLON:
		tidx += 1
		if tidx < len(token_list) and token_list[tidx][0] != NUMBER:
			raise Exception, 'Expected a number to follow a colon'
			
		distance = token_list[tidx][1]
		tidx += 1
		
	return (node,distance,children), tidx

#####
# Parsing code...
	
COMMA = ','
COLON = ':'
SEMICOLON = ';'
LPAREN = '('
RPAREN = ')'
QUOTE = '"'
EQUALS = '='
STRING = 'string'
NUMBER = 'number'

def make_token_list(fh):
	"""
	Read all tokens up until a semi-colon is read or the end of file is reached.
	"""
	token_list = []
	next_char = None
	c = None
	
	while True:

		# get the next character
		if next_char:
			c,next_char = next_char,None
		else:
			c = fh.read(1)
		
		# skip whitespace
		if len(c) == 0:
			break
		elif c.isspace():
			continue
		elif c == QUOTE:
			string = read_quoted_string(fh)
			token_list.append((STRING,string))
		elif c == COLON or c == LPAREN or c == RPAREN or c == EQUALS or c == COMMA:
			token_list.append((c,c))
		elif c == SEMICOLON:
			token_list.append((c,c))
		elif c.isalpha():
			next_char,string = read_unquoted_string(fh,c)
			token_list.append((STRING,string))
		elif c.isdigit():
			next_char,num = read_number(fh,c)
			token_list.append((NUMBER,num))
		else:
			raise Exception, 'Unknown character encountered: %s' % c
		
	return token_list
		
def read_quoted_string(fh):
	string = ''
	c = fh.read(1)
	
	while c != '"':
		string += c
		c = fh.read(1)
		
		if len(c) == 0:
			raise Exception, 'Unterminated quoted string beginning with "%s"' % string
		
	return string
	
def read_unquoted_string(fh,c):
	string = c
	
	c = fh.read(1)
	while len(c) > 0 and not is_reserved(c) and not c.isspace():
		string += c
		c = fh.read(1)
		
	return c, string
		
def read_number(fh,c):
	num_str = c
	
	c = fh.read(1)
	while len(c) > 0 and not is_reserved(c) and not c.isspace():
		num_str += c
		c = fh.read(1)
	
	num = None
	if num_str.find('.') == -1:
		num = int(num_str)
	else:
		num = float(num_str)	
	
	return c, num
		
def is_reserved(c):
	return c == COLON or c == LPAREN or c == RPAREN or c == EQUALS or c == COMMA or c == SEMICOLON
		
def main():
	pass


if __name__ == '__main__':
	main()

