"""
This module provides implementations of parsimony measures for trees.
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

__all__ = ['compute_parsimony_score']

def compute_parsimony_score(T,leaf_vectors):
	"""
	Compute the parsimony score for the leaf vectors on the tree T.  leaf_vectors is a
	dictionary mapping all the leaf nodes to a list of objects.  Each leaf must have
	a vector of the same length.
	"""
	
	vlen = len(leaf_vectors.itervalues().next())
	
	#####
	# Compute the pscore for each character
	pscore = 0
	for i in range(0,vlen):
		potential_assignments = {}
		__compute_potential_assignments(T,T.root(),i,leaf_vectors,potential_assignments)
		assignments = {}
		__select_assignments(T,T.root(),potential_assignments,assignments,None)
		pscore += __compute_pscore(T,T.root(),assignments)
		
	return pscore

def __compute_pscore(T,node,assignments):
	pscore = 0
	
	for c in T.children_iter(node):
		if assignments[node] != assignments[c]:
			pscore += 1
		
		pscore += __compute_pscore(T,c,assignments)
		
	return pscore

def __select_assignments(T,node,pa,assignments,parent_assignment):
	comb = pa[node]

	if len(comb) == 0:
		assignments[node] = parent_assignment
	if len(comb) == 1:
		assignments[node] = comb.pop()
	elif len(comb) > 1:
		if parent_assignment is not None and parent_assignment in comb:
			assignments[node] = parent_assignment
		else:
			assignments[node] = comb.pop()

	# update all children
	for c in T.children_iter(node):
		__select_assignments(T,c,pa,assignments,assignments[node])
		
	return

def __compute_potential_assignments(T,node,i,leaf_vectors,pa):
	comb = set()
	if T.is_leaf(node):
		comb.add(leaf_vectors[node][i])
	else:
		all_choices = []
		for c in T.children_iter(node):
			all_choices.extend(__compute_potential_assignments(T,c,i,leaf_vectors,pa))
		
		# only keep those values that have a maximum representation in the children
		counts = [all_choices.count(x) for x in all_choices]
		max_counts = max(counts)
		max_idxs = [i for i in range(len(counts)) if counts[i] == max_counts]
		comb.update([all_choices[i] for i in max_idxs])
			
	pa[node] = comb	
	return comb

def main():
	pass


if __name__ == '__main__':
	main()

