"""
This module encapsulates functions for working with paths, in particular for computing
path weights.

A path in PyGraph is a list of edges, L, such that for every e1 = L[i], e2 = L[i+1]:

	- In an undirected graph e1 and e2 share an endpoint in common
	- In a directed graph, target(e1) == source(e2)
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

import types

__all__ = [	'get_weight','get_weight_by_dict','get_weight_by_fxn','build_weight_comparator',
			'get_node_list','enum_simple_paths']

def build_weight_comparator(weights):
	"""
	Returns a comparator function that uses the weights of input paths to determine
	the comparison outcome.  The value returned by the comparator for inputs (p1,p2) 
	returns values (-1,0,+1) which conform to the conventions used by comparators.
	"""
	def compare(p1,p2):
		w1 = get_weight(p1,weights)
		w2 = get_weight(p2,weights)
		
		if w1 < w2:
			return -1
		elif w1 == w2:
			return 0
		else:
			return 1
			
	return compare

def get_weight_by_dict(path,weights):
	"""
	Compute the weight of the path.  weights is a dictionary such that weights[e] returns the 
	weight of edge e for each edge in the path.
	"""
	weight = 0
	for e in path:
		weight += weights[e]
		
	return weight

def get_weight_by_fxn(path,weights):
	"""
	Compute the weight of the path.  weights is a function such that weights(e) returns the 
	weight of edge e for each edge in the path.
	"""
	weight = 0
	for e in path:
		weight += weights(e)
		
	return weight
	
def get_weight(path,weights):
	"""
	Compute the weight of the path.  weights is either a dictionary or a unary function that maps
	a weight to each edge in the path.
	"""
	if type(weights) == types.DictionaryType:
		return get_weight_by_dict(path,weights)
	elif type(weights) == types.FunctionType:
		return get_weight_by_fxn(path,weights)
	else:
		raise PyGraphException, 'weights must be either a dictionary or unary function'

def get_node_list(G,path):
	"""
	Return the list of nodes which constitutes this path.  In directed graphs, the node ordering is 
	unique.  In undirected graphs, however, the ordering of nodes in a short two edge self-loop 
	cannot be determined.  In this situation, this method will order nodes using its best guess.
	"""
	nlist = None
	
	if G.is_directed():
		if len(path) > 0:
			nlist = [G.source(path[0])]
		else:
			nlist = []
			
		for e in path:
			nlist.append(G.target(e))
		
		return nlist
	else:
		if len(path) == 0:
			return []
		
		if len(path) == 1:
			x,y = G.endpoints(path[0])
			return [x,y]
		
		else:
			a,b = G.endpoints(path[0])
			c,d = G.endpoints(path[1])
			
			if a == c:
				a,b = b,a
				c,d = d,c
			elif b != d:
				a,b = b,a
				
			if b != d:
				raise PyGraphException,'Endpoints of path do not match up'
			
			nlist = [a]
	
			for e in path:
				x,y = G.endpoints(e)
				if x != nlist[-1]:
					x,y = y,x
					
				nlist.append(y)
		
			return nlist

#####
# Simple path enumeration
def enum_simple_paths(G,srcs,sinks,path_list=None):
	"""
	This function finds all simple paths (paths without loops) in graph G connecting nodes in srcs
	to any of the nodes in sinks.  srcs is an iterable of nodes.  sinks is a set of nodes.  The 
	path_list is a list to which all paths found are added as they are discovered.
	
	If path_list is not set, then a list of the paths found are returned.
	"""
	
	ret_pl = False
	
	if path_list is None:
		ret_pl = True
		path_list = []
	for s in srcs:
		__enumerate_simple_paths(G,s,sinks,[],set(),{},path_list)
		
	if ret_pl:
		return path_list

def __enumerate_simple_paths(graph, node, outputs, curr_path, curr_path_set, cached_paths, path_list):
    
	if node in outputs:
		path_list.append(list(curr_path))

		# update cached paths
		last_node = node
		ns = set()
		ns.add(node)
		for k in range(len(curr_path)-1,1):
			e = curr_path[k]
			n = graph.endpoint(e,last_node)
			ns.add(n)
			
			if n not in cached_paths:
				cached_paths[n] = []
				
			p = curr_path[k:]
			cached_paths[n].append((p,set(ns)))
			last_node = n
			
		return
        
	if node in curr_path_set:
		return
    
	if node in cached_paths:
		for path, pset in cached_paths[node]:
			# watch for loops
			if len(pset.intersection(curr_path_set)) == 0:
				path_list.append(curr_path + path)
                
		return
    
	curr_path_set.add(node)
    
	for e in graph.E.iter(node):
		curr_path.append(e)
		__enumerate_simple_paths(graph, graph.target(e), outputs, curr_path, curr_path_set, cached_paths, path_list)
		del curr_path[-1]
        
	curr_path_set.remove(node)
    
	return
