"""
This module provides functions for assessing properties of graphs.
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy

__all__ = ['degree_dist', 'cluster_coeff']

def degree_dist(G,normalize=True):
    """
    Return the degree distribution of graph G.  The result, D, is a numpy array with 
    length G.degree() + 1.  D[i] is the fraction of nodes with degree i.  
    If normalize is False, then D[i] is the number of nodes in the graph with degree i.
    
    Currently only the undirected degree distribution is supported.
    """
    
    max_degree = G.degree()
    X = numpy.zeros((max_degree+1,))
    for n in G.V:
        if G.degree(n) > 0:
            X[G.degree(n)] += 1
            
    if normalize:
        num_nodes = float(len(G.V))
        X = X / num_nodes
        
    return X

def cluster_coeff(G):
    """
    Compute the clustering coefficient for the graph specified.  If the graph is undirected,
    then the undirected clustering coefficient is computed.  If the graph is directed, then
    the directed clustering coefficient is computed.
    """
    edge_lookup = set()
    
    cc = 0.
    
    if G.is_directed():
        for e in G.E:
            u,v = G.endpoints(e)
            edge_lookup.add((u,v))
            
        for v in G.V:
            N = G.neighbors(v)
            ecount = 0
            for n1, ni in zip(N,range(len(N))):
                for nj in range(ni+1,len(N)):
                    n2 = N[nj]
                    if (n1,n2) in edge_lookup:
                        ecount += 1
                    if (n2,n1) in edge_lookup:
                        ecount += 1
        
            if len(N) == 1:
                cc += 1.0
            if len(N) > 1:
                cc += float(ecount) / float(len(N) * (len(N) - 1))
        
        cc = cc / float(len(G.V))            
    else:
        for e in G.E:
            u,v = G.endpoints(e)
            edge_lookup.add((u,v))
            edge_lookup.add((v,u))
    
        for v in G.V:
            N = G.neighbors(v)
            ecount = 0
            for n1, ni in zip(N,range(len(N))):
                for nj in range(ni+1,len(N)):
                    n2 = N[nj]
                    if (n1,n2) in edge_lookup:
                        ecount += 1
        
            if len(N) == 1:
                cc += 1.0
            if len(N) > 1:
                cc += float(2 * ecount) / float(len(N) * (len(N) - 1))
        
        cc = cc / float(len(G.V))
        
    return cc        