"""
This module defines classes, functions, and constants useful in computing shortest and near-shortest paths
in graphs.
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

import types
import array
import pyweaver.util.minheap as minheap
import path as path_utils

# the set of functions we export
__all__ = ['dijkstra','dijkstra_weight','floyd_warshall','carlyle']

####
# Classes
class DijkstraHeapNodeHolder:
	"""
	This class is used by the Dijkstra implementations in order to hold nodes in the heap.
	This is required because the heap must be able to compare nodes by their distances to
	the source node.
	"""
	
	def __init__(self,n,distances):
		self.node = n
		self.distances = distances
	
	def __cmp__(self,other):
		"""
		self < other if self.node is closer from source
		self == other if self.node and other.node are equidistant
		self > other if self.node is further from source
		"""
		dself = self.distances[self.node]
		dn = self.distances[other.node]
		if dself < dn:
			return -1
		elif dself == dn:
			return 0
		else:
			return 1

####
# Functions
def floyd_warshall(G,weights=lambda x: 1):
	"""
	Compute all-pairs shortest paths in graph G using the Floyd-Warshall algorithm.  This method
	returns a tuple (distances,preds,nidx_lookup) where distances[i][j] is the weight of the shortest path
	connecting node with index i to node with index j and preds is a matrix of the predecessors 
	comprising the shortest paths.  The index for node n is returned by nidx_lookup[n].
	
	If weights is specified, it must be a function or a dictionary
	that evaluates to the weight of an edge in graph G. By default it returns a weight
	of one for every edge in the graph.
	"""
	
	# setup the distances and preds matrices
	num_nodes = len(G.N)
	init_list = [float('infinity')] * num_nodes
	distances = [array.array('f',init_list) for i in range(0,num_nodes)]
	
	preds = [[None] * num_nodes for i in range(0,num_nodes)]
	
	# get a listing of nodes
	nlist = G.N()
	nidx_lookup = {}
	for n,i in zip(nlist,range(0,len(nlist))):
		nidx_lookup[n] = i
		       
	# initialize the distances matrix	
	is_dict = type(weights) == types.DictionaryType
	is_directed = G.is_directed()
	for e in G.E:
		w = None
		if is_dict:
			w = weights[e]
		else:
			w = weights(e)
			
		x,y = G.endpoints(e)
		xidx = nidx_lookup[x]; yidx = nidx_lookup[y]

		distances[xidx][yidx] = w		
		if not is_directed:
			distances[yidx][xidx] = w
		
	# perform dynamic programming part of algorithm
	for midx in range(0,len(nlist)):
		mid_node = nlist[midx]
		
		for i in range(0,len(nlist)):
			for j in range(0,len(nlist)):
				d = distances[i][j]
				d2 = distances[i][midx] + distances[midx][j]
		
				if d2 < d:
					preds[i][j] = mid_node
					distances[i][j] = d2
	
	return (distances,preds,nidx_lookup)
		
def inner_run_dijkstra(G,weights,source,reverse=True):
	"""
	This function is a helper function for the public dijkstra methods.  It does the
	heavy-lifting for the Dijkstra algorithm.  It returns a triple of dictionaries
	(distance,pred,weight_lookup) in which distance[n] is the weight of the shortest path 
	from source to n.  pred[n] is the pair (u,e) specifying that u is the predecessor node 
	in the shortest path connectins source to n and e is the edge that connects u to n.
	weight_lookup[e] is the weight of edge e.
	
	weights can be either a dictionary or a unary function.  As a dictionary, weights[e]
	must be defined for all edges e in G.  As a function, weights(e) must return a number
	for every edge e in G.
	
	If reverse it True, then dijkstra will be run BACKWARDS on the graph.  This won't make a
	difference for undirected graphs.  However, on directed graphs, all edges will be traversed
	backwards.  This is useful when trying to compute every node's shortest path to a specific
	target.
	"""
	# initialize distances
	distance = {}
	pred = {}
	for n in G.N:
		distance[n] = float('infinity')
		pred[n] = None
	
	distance[source] = 0
	
	# setup weight_lookup
	weight_lookup = None
	if type(weights) == types.DictionaryType:
		weight_lookup = weights
	elif type(weights) == types.FunctionType:
		weight_lookup = {}
		for e in G.E:
			weight_lookup[e] = weights(e)
	else:
		raise Exception, 'weights must be either a dictionary or unary function'
	
	# populate heap
	heap = [DijkstraHeapNodeHolder(n,distance) for n in G.N]
	heap_lookup = {}
	for obj in heap:
		heap_lookup[obj.node] = obj

	minheap.heapify(heap)
	
	# run dijkstra's algorithm
	while len(heap) > 0:
		#u = heapq.heappop(heap).node
		u = minheap.pop(heap).node

		edges = None
		if not reverse:
			edges = G.E(u)
		else:
			edges = G.E(v=u)

		for e in edges:
			v = G.endpoint(e,u)
			new_dist = distance[u] + weight_lookup[e]
			if distance[v] > new_dist:
				#print '%f vs %f' % (distance[v],new_dist)
				distance[v] = new_dist
				pred[v] = (u,e)
				
				# update heap
				minheap.decrease_key(heap, heap_lookup[v])
				
	return (distance,pred,weight_lookup)

def dijkstra(G,weights=lambda x: 1,source=None,target=None):
	"""
	Uses Dijstra's algorithm to compute a shortest path from the source to the target.
	
	The weights value is either a dictionary or a function.  As a dictionary, weights[e]
	must be defined for all edges e in G.  As a function, weights(e) must return a number
	for every edge e in G.  Note that Dijkstra's algorithm is only guaranteed to work if 
	all weights are non-negative.  The default weights parameter is a function that returns
	a weight of 1 for every edge.
	
	A list of the edges on the path from the source to the target is returned.
	
	If only source is specified, then the single source shortest paths are computed from the source
	to all nodes in the graph.  In this case, a dictionary is returned in which the key is
	the target node and the value is the list of edges comprising the path.
	
	If only target is specified, then the single shortest paths are computed from each node in the 
	graph to the specified target.  In this case, a dictionary is returned in which the key is
	the source node and the value is the list of edges comprising the path.
	"""
	if source is None and target is None:
		raise Exception, 'source and target cannot both be None'
	
	# run dijkstra's algorithm
	distance = None; pred = None; weight_lookup = None
	if source:
		distance, pred, weight_lookup = inner_run_dijkstra(G,weights,source,reverse=False)
	else:
		distance, pred, weight_lookup = inner_run_dijkstra(G,weights,target,reverse=True)
	
	# build paths
	if source and target: # Compute the single shortest path to target
		path = []
		last = target
		while last != source:
			pair = pred[last]
			
			# there's no predecessor...
			if pair is None:
				path = None
				break

			last,e = pair
			
			path.insert(0,e)
		
		return path
	elif source is None: # Compute the shortest path from each node to target
		paths = {}
		
		for n in G.N:
			if n == source:
				continue
		
			path = []
			last = n
			while last != target:
				pair = pred[last]
				
				# there's no predecessor...
				if pair is None:
					path = None
					break

				last,e = pair	
				
				path.insert(0,e)
			
			if path:
				path.reverse()
				
			paths[n] = path
		
		return paths
		
	else: # Compute the shortest path from source to each node
		paths = {}
		
		for n in G.N:
			if n == source:
				continue
		
			path = []
			last = n
			while last != source:
				pair = pred[last]
				
				# there's no predecessor...
				if pair is None:
					path = None
					break

				last,e = pair	
				
				path.insert(0,e)
				
			paths[n] = path
			
		return paths
		
def dijkstra_weight(G,weights=lambda x: 1,source=None,target=None):
	"""
	Uses Dijstra's algorithm to compute the weight of the shortest path from the source 
	to the target.
	
	The weights value is either a dictionary or a function.  As a dictionary, weights[e]
	must be defined for all edges e in G.  As a function, weights(e) must return a number
	for every edge e in G.  Note that Dijkstra's algorithm is only guaranteed to work if 
	all weights are non-negative.  The default weights parameter is a function that returns
	a weight of 1 for every edge.
	
	The numeric weight of the path from the source to the target is returned.  If no path
	could be found, infinity is returned (= float('infinity') ).
	
	If only source is specified, then the single source shortest paths are computed from the source
	to all nodes in the graph.  In this case, a dictionary is returned in which the key is
	the target node and the value is the weight of the edge connecting path.
	
	If only target is specified, then the single shortest paths are computed from each node in the 
	graph to the specified target.  In this case, a dictionary is returned in which the key is
	the source node and the value is the list of edges comprising the path.
	"""
	if source is None and target is None:
		raise Exception, 'source and target cannot both be None'
	
	# run dijkstra's algorithm
	distance = None; pred = None; weight_lookup = None
	if source:
		distance, pred, weight_lookup = inner_run_dijkstra(G,weights,source,reverse=False)
	else:
		distance, pred, weight_lookup = inner_run_dijkstra(G,weights,target,reverse=True)
	
	# build paths
	if source and target: # Compute the single shortest path to target
		weight = 0
		last = target
		while last != source:
			pair = pred[last]
			
			# there's no predecessor...
			if pair is None:
				weight = float('infinity')
				break
				
			last,e = pair
			
			weight += weight_lookup[e]
		
		return weight
	
	elif source is None: # Compute the shortest path from each node to the target
		all_weights = {}
		
		for n in G.N:	
			if n == source:
				continue
		
			weight = 0
			last = n
			while last != target:
				pair = pred[last]
				
				# there's no predecessor...
				if pair is None:
					weight = float('infinity')
					break
					
				last,e = pair
				
				weight += weight_lookup[e]
			
			all_weights[n] = weight
			
		return all_weights
	
	else: # Compute the shortest path from source to each node
		
		all_weights = {}
		
		for n in G.N:	
			if n == source:
				continue
		
			weight = 0
			last = n
			while last != source:
				pair = pred[last]
				
				# there's no predecessor...
				if pair is None:
					weight = float('infinity')
					break
					
				last,e = pair
				
				weight += weight_lookup[e]
			
			all_weights[n] = weight
			
		return all_weights
		
def carlyle(G,weights,source,target,epsilon):
	"""
	Computes the simple paths from u to v with weights greater than the shortest path by at 
	most a factor of epsilon.  The weights value is either a dictionary or a function.  As a dictionary, weights[e]
	must be defined for all edges e in G.  As a function, weights(e) must return a number
	for every edge e in G.  Note that all weights must be non-negative.
	
	A list of paths are returned. The paths are in order of shortest to longest.
	
	This algorithm was taken from the 2003 paper entitled 
	"Near-Shortest and K-Shortest Simple Paths" by W. M. Carlyle.
	"""
	# convert weights to a dictionary
	weight_lookup = None
	if type(weights) == types.DictionaryType:
		weight_lookup = weights
	elif type(weights) == types.FunctionType:
		weight_lookup = {}
		for e in G.edges():
			weight_lookup[e] = weights(e)
	else:
		raise PyWeaverException, 'weights must be either a dictionary or unary function'
	
	####
	# algorithm starts here...
	all_weights = dijkstra_weight(G,weight_lookup,target=target)
	max_weight = (1. + epsilon) * float(all_weights[source])
	
	# we will use this for 'removing' edges from the graph without having to change
	# its topology.
	infinity = float('infinity')
	tmp_weight_lookup = dict(weight_lookup)
	
	paths = []
	
	path_weight = 0
	path = []
	stack = [source]
	iterators = [G.E.iter(source)]
	while len(stack) > 0:
		u = stack[-1]
		it = iterators[-1]
		
		edge = None
		try:
			edge = it.next()
		except StopIteration:
			pass
		
		if edge:
			v = G.endpoint(edge,u)
			dist = path_weight + weight_lookup[edge] + all_weights[v]
			if v not in stack and dist <= max_weight:
				if v == target:
					path.append(edge)
					paths.append(list(path))
					path.pop()
				else:
					stack.append(v)
					iterators.append(G.E.iter(v))
					path.append(edge)
					path_weight += weight_lookup[edge]
					
					# recompute the shortest paths to target excluding every node in the path
					# here we remove the most recent addition to the path by making the edge's
					# weight infinity (totally untraversable
					tmp_weight_lookup[edge] = infinity
					all_weights = dijkstra_weight(G,tmp_weight_lookup,target=target)
		else:
			u = stack.pop()
			iterators.pop()
			
			if len(path) > 0:
				edge = path.pop()
				path_weight -= weight_lookup[edge]
				tmp_weight_lookup[edge] = weight_lookup[edge]
			
			# recompute the shortest paths to target.  Now this edge can be part of
			# shortest paths
			all_weights = dijkstra_weight(G,tmp_weight_lookup,target=target)
	
	# order paths by weight
	paths.sort(path_utils.build_weight_comparator(weight_lookup))
	
	return paths