"""
Dijstra test functions.
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest
import pyweaver.graph as graph
from pyweaver.shortestpath import dijkstra, dijkstra_weight

class DijkstraTests(unittest.TestCase):
	
	def test_sp_undirected(self):
		# make graph		
		G = graph.Graph()
		
		G.E.add(1,2,'e1')
		G.E.add(2,4,'e2')
		G.E.add(1,3,'e3')
		G.E.add(4,5,'e4')
		G.E.add(3,5,'e5')
		G.N.add(6)
		
		# set weights
		weights = {'e1':1., 'e2':2., 'e3':2, 'e4':0.5, 'e5':3}
		
		# find single shortest paths
		path = dijkstra(G,weights,1,5)
		solution = ['e1','e2','e4']
		
		self.assertEqual(path,solution,'Single shortest path failed.')
		
		# find all shortest paths
		paths = dijkstra(G,weights,1)
		solution2 = ['e1']
		solution3 = ['e3']
		solution4 = ['e1','e2']
		solution5 = ['e1','e2','e4']
		solution6 = None
		
		self.assertEqual(paths[2],solution2,'Shortest path failed.')
		self.assertEqual(paths[3],solution3,'Shortest path failed.')
		self.assertEqual(paths[4],solution4,'Shortest path failed.')
		self.assertEqual(paths[5],solution5,'Shortest path failed.')
		self.assertEqual(paths[6],solution6,'Shortest path failed.')
		
	def test_sp_directed(self):
		# make graph		
		G = graph.DiGraph()
		
		G.E.add(1,2,'e1')
		G.E.add(2,4,'e2')
		G.E.add(1,3,'e3')
		G.E.add(4,5,'e4')
		G.E.add(3,5,'e5')
		G.N.add(6)
		
		# set weights
		weights = {'e1':1., 'e2':2., 'e3':2, 'e4':0.5, 'e5':3}
		
		# find single shortest paths
		path = dijkstra(G,weights,1,5)
		solution = ['e1','e2','e4']
		
		self.assertEqual(path,solution,'Single shortest path failed.')
		
		# find all shortest paths
		paths = dijkstra(G,weights,1)
		solution2 = ['e1']
		solution3 = ['e3']
		solution4 = ['e1','e2']
		solution5 = ['e1','e2','e4']
		solution6 = None
		
		self.assertEqual(paths[2],solution2,'Shortest path failed.')
		self.assertEqual(paths[3],solution3,'Shortest path failed.')
		self.assertEqual(paths[4],solution4,'Shortest path failed.')
		self.assertEqual(paths[5],solution5,'Shortest path failed.')
		self.assertEqual(paths[6],solution6,'Shortest path failed.')
		
	def test_spw_undirected(self):
		"""
		Test dikjstra_weight on an undirected graph
		"""
		# make graph		
		G = graph.Graph()
		
		G.E.add(1,2,'e1')
		G.E.add(2,4,'e2')
		G.E.add(1,3,'e3')
		G.E.add(4,5,'e4')
		G.E.add(3,5,'e5')
		G.N.add(6)
		
		# set weights
		weights = {'e1':1., 'e2':2., 'e3':2, 'e4':0.5, 'e5':3}
		
		# find single shortest paths
		weight = dijkstra_weight(G,weights,1,5)
		solution = 3.5
		
		self.assertEqual(weight,solution,'Single shortest path failed.')
		
		# find all shortest paths
		all_weights = dijkstra_weight(G,weights,1)
		solution2 = 1
		solution3 = 2
		solution4 = 3
		solution5 = 3.5
		solution6 = float('infinity')
		
		self.assertEqual(all_weights[2],solution2,'Shortest path weights failed.')
		self.assertEqual(all_weights[3],solution3,'Shortest path weights failed.')
		self.assertEqual(all_weights[4],solution4,'Shortest path weights failed.')
		self.assertEqual(all_weights[5],solution5,'Shortest path weights failed.')
		self.assertEqual(all_weights[6],solution6,'Shortest path weights failed.')
		
	def test_spw_directed(self):
		"""
		Test dikjstra_weight on a directed graph
		"""	
		# make graph		
		G = graph.DiGraph()
		
		G.E.add(1,2,'e1')
		G.E.add(2,4,'e2')
		G.E.add(1,3,'e3')
		G.E.add(4,5,'e4')
		G.E.add(3,5,'e5')
		G.N.add(6)
		
		# set weights
		weights = {'e1':1., 'e2':2., 'e3':2, 'e4':0.5, 'e5':3}
		
		# find single shortest paths
		weight = dijkstra_weight(G,weights,1,5)
		solution = 3.5
		
		self.assertEqual(weight,solution,'Single shortest path failed.')
		
		# find all shortest paths
		all_weights = dijkstra_weight(G,weights,1)
		solution2 = 1
		solution3 = 2
		solution4 = 3
		solution5 = 3.5
		solution6 = float('infinity')
		
		self.assertEqual(all_weights[2],solution2,'Shortest path weights failed.')
		self.assertEqual(all_weights[3],solution3,'Shortest path weights failed.')
		self.assertEqual(all_weights[4],solution4,'Shortest path weights failed.')
		self.assertEqual(all_weights[5],solution5,'Shortest path weights failed.')
		self.assertEqual(all_weights[6],solution6,'Shortest path weights failed.')
	
	###########
	# BELOW WE ARE TESTING THE REVERSE DIJKSTRA
	###########
	
	def test_sp_undirected_reverse(self):
		# make graph		
		G = graph.Graph()
		
		G.E.add(1,2,'e1')
		G.E.add(2,4,'e2')
		G.E.add(1,3,'e3')
		G.E.add(4,5,'e4')
		G.E.add(3,5,'e5')
		G.N.add(6)
		
		# set weights
		weights = {'e1':1., 'e2':2., 'e3':2, 'e4':0.5, 'e5':3}
		
		# find all shortest paths
		paths = dijkstra(G,weights,target=5)
		solution1 = ['e1','e2','e4']
		solution2 = ['e2','e4']
		solution3 = ['e5']
		solution4 = ['e4']
		solution6 = None

		self.assertEqual(paths[1],solution1,'Shortest path failed.')		
		self.assertEqual(paths[2],solution2,'Shortest path failed.')
		self.assertEqual(paths[3],solution3,'Shortest path failed.')
		self.assertEqual(paths[4],solution4,'Shortest path failed.')
		self.assertEqual(paths[6],solution6,'Shortest path failed.')
		
	def test_sp_directed_reverse(self):
		# make graph		
		G = graph.DiGraph()
		
		G.E.add(1,2,'e1')
		G.E.add(2,4,'e2')
		G.E.add(1,3,'e3')
		G.E.add(4,5,'e4')
		G.E.add(3,5,'e5')
		G.N.add(6)
		
		# set weights
		weights = {'e1':1., 'e2':2., 'e3':2, 'e4':0.5, 'e5':3}
		
		# find all shortest paths
		paths = dijkstra(G,weights,target=5)
		solution1 = ['e1','e2','e4']
		solution2 = ['e2','e4']
		solution3 = ['e5']
		solution4 = ['e4']
		solution6 = None

		self.assertEqual(paths[1],solution1,'Shortest path failed.')		
		self.assertEqual(paths[2],solution2,'Shortest path failed.')
		self.assertEqual(paths[3],solution3,'Shortest path failed.')
		self.assertEqual(paths[4],solution4,'Shortest path failed.')
		self.assertEqual(paths[6],solution6,'Shortest path failed.')
		
	def test_spw_undirected_reverse(self):
		"""
		Test dikjstra_weight on an undirected graph
		"""
		# make graph		
		G = graph.Graph()
		
		G.E.add(1,2,'e1')
		G.E.add(2,4,'e2')
		G.E.add(1,3,'e3')
		G.E.add(4,5,'e4')
		G.E.add(3,5,'e5')
		G.N.add(6)
		
		# set weights
		weights = {'e1':1., 'e2':2., 'e3':2, 'e4':0.5, 'e5':3}
		
		# find all shortest paths
		all_weights = dijkstra_weight(G,weights,target=5)
		solution1 = 3.5
		solution2 = 2.5
		solution3 = 3
		solution4 = 0.5
		solution6 = float('infinity')
		
		self.assertEqual(all_weights[1],solution1,'Shortest path weights failed.')
		self.assertEqual(all_weights[2],solution2,'Shortest path weights failed.')
		self.assertEqual(all_weights[3],solution3,'Shortest path weights failed.')
		self.assertEqual(all_weights[4],solution4,'Shortest path weights failed.')
		self.assertEqual(all_weights[6],solution6,'Shortest path weights failed.')
		
	def test_spw_directed_reverse(self):
		"""
		Test dikjstra_weight on a directed graph
		"""	
		# make graph		
		G = graph.DiGraph()
		
		G.E.add(1,2,'e1')
		G.E.add(2,4,'e2')
		G.E.add(1,3,'e3')
		G.E.add(4,5,'e4')
		G.E.add(3,5,'e5')
		G.N.add(6)
		
		# set weights
		weights = {'e1':1., 'e2':2., 'e3':2, 'e4':0.5, 'e5':3}
		
		# find all shortest paths
		all_weights = dijkstra_weight(G,weights,target=5)
		solution1 = 3.5
		solution2 = 2.5
		solution3 = 3
		solution4 = 0.5
		solution6 = float('infinity')
		
		self.assertEqual(all_weights[1],solution1,'Shortest path weights failed.')
		self.assertEqual(all_weights[2],solution2,'Shortest path weights failed.')
		self.assertEqual(all_weights[3],solution3,'Shortest path weights failed.')
		self.assertEqual(all_weights[4],solution4,'Shortest path weights failed.')
		self.assertEqual(all_weights[6],solution6,'Shortest path weights failed.')
		
if __name__ == '__main__':
	unittest.main()