"""
Unit tests for the undirected graph.
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest
import pyweaver.graph as graph

class UnDirectedGraphTests(unittest.TestCase):

	def test_endpoints(self):
		"""
		Test the methods that retrieve endpoints.
		"""
		G = graph.Graph()
		G.E.add(1,2,'e1')
		G.E.add(1,3,'e2')
		G.E.add(3,4,'e3')
		G.E.add(4,5,'e4')
		
		G.N.add(6)
		
		x,y = G.endpoints('e1')
		self.assertTrue( (x == 1 and y == 2) or (x == 2 and y == 1) )
		
		x = G.endpoint('e2',1)
		self.assertEqual(x,3)

	def test_num_methods(self):
		"""
		Test the methods that keep count of the number of edges/nodes.
		"""
		G = graph.Graph()
		G.E.add(1,2,'e1')
		G.E.add(1,3,'e2')
		G.E.add(3,4,'e3')
		G.E.add(4,5,'e4')
		
		G.N.add(6)
		
		# The basic forms
		self.assertEqual(len(G.N),6,'Incorrect node count')
		self.assertEqual(len(G.E),4,'Incorrect edge count')
		
		# The connectivity forms
		self.assertEqual(len(G.E(u=4)),2,'Incorrect edge count')
		self.assertEqual(len(G.E(v=4)),2,'Incorrect edge count')
		
		self.assertEqual(len(G.E(2,2)),0,'Incorrect self-loop count')
		self.assertEqual(len(G.E(6)),0,'Incorrect self-loop count')
		self.assertEqual(len(G.E(5,6)),0,'Incorrect self-loop count')
		
		# remove some nodes and edges
		G.N.remove(4)
		G.E.remove('e2')
		
		self.assertEqual(len(G.E(3)),0,'Incorrect edge count')
		self.assertEqual(len(G.E(1)),1,'Incorrect edge count')
		
	def test_neighbors(self):

		G = graph.Graph()
		G.E.add(1,2)
		G.E.add(1,3)
		G.E.add(1,4)
		G.E.add(2,3)

		self.assertEqual(len(G.neighbors(1)),3)
		self.assertEqual(len(G.neighbors(2)),2)
		self.assertEqual(len(G.neighbors(4)),1)
		
	def test_index(self):
		"""
		Test the methods that keep count of the number of edges/nodes.
		"""
		G = graph.Graph()
		G.N.index('x')
		G.E.index('g')
		n1 = graph.Node(1, x = 1)
		n2 = graph.Node(2, x = 2)
		n3 = graph.Node(3, x = 2)
		n4 = graph.Node(4, x = 2)
		n5 = graph.Node(5, x = 3)
		n6 = graph.Node(6, x = 3)
		
		e1 = graph.Edge(g = 1)
		e2 = graph.Edge(g = 1)
		e3 = graph.Edge(g = 1)
		e4 = graph.Edge(g = 2)
		
		G.E.add(n1,n2,e1)
		G.E.add(n1,n3,e2)
		G.E.add(n3,n4,e3)
		G.E.add(n4,n5,e4)
		
		G.N.add(n6)		

		# test node indexing
		self.assertEquals(G.V[1],n1)
		self.assertEquals(len(G.V.find_all(x=2)),3)
		self.assertEquals(set(G.V.find_all(x=3)), set([n5,n6]))
		
		n1.x = 4
		n6.x = 2
		
		G.V.reindex([n1,n6])
		
		self.assertEquals(G.V[1],None)
		self.assertEquals(G.V[4],n1)
		self.assertEquals(len(G.V.find_all(x=2)),4)
		self.assertEquals(set(G.V.find_all(x=3)), set([n5]))
		
		# test edge indexing
		self.assertEquals(G.E[2],e4)
		self.assertEquals(len(G.E.find_all(g=1)),3)
		
		e1.x = 2; e2.x = 2
		G.E.reindex([e1,e2])
		
		self.assertEquals(G.E[2],e4)
		self.assertEquals(len(G.E.find_all(g=1)),3)
		
		e1.g = 2; e2.g = 2
		G.E.reindex([e1,e2])
		
		self.assertEquals(len(G.E.find_all(g=1)),1)
		self.assertEquals(set(G.E.find_all(g=2)),set([e1,e2,e4]))

####
# Code to run all the tests
if __name__ == '__main__':
	unittest.main()