"""
This module defines the tree data structures: unrooted tree (UTree) and rooted tree (RTree).

These are both specializations of general graphs.  The unrooted tree is an undirected graph in
which there is only one path between any two nodes.  The rooted tree is a directed graph in
which the edges eminate from the root towards the leaves and there is only one path between any
two nodes.

The unrooted tree is not yet implemented.
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

from graph import *

__all__ = ['RTree','Node','Edge']

class RTree(DiGraph):
	"""
	The rooted tree is a specialized directed graph.  In order to ensure that its structure is
	maintained, the structure should only be manipulated through the tree methods:
	
		- add_child(...)
		- remove_child(...)
		- move(...)
		
	All graph methods can be used for accessing and interrogating the graph structure.
	"""
	
	def __init__(self,root,edge_generator=lambda p,c,x: Edge(length=x),edge_length_fxn=lambda e: e.length):
		"""
		As this tree is rooted, a root must be specified.
		
		The type of edge decorators used in this tree can be specified by providing an
		edge generator function, edge_generator, which accepts the parent node, the child node, and
		the the branch length (which may be None, for no length), and returns an edge decorator.  
		By default, edges of type Edge are created.
		
		For custom edge types, it may be necessary to specify an edge length function that returns
		the length of the edge passed in.  By default, this function simply returns the value of
		the 'length' field.
		"""
		DiGraph.__init__(self)
		
		self.root_node = root
		self.N.add(root)
		
		# setup the edge generator
		self.gen_edge = edge_generator
		self.get_edge_length = edge_length_fxn

		# setup internal tree structure - optimized for speed
		self.parent_lookup = {root : None}
		self.children_lookup = {root : []}
		
	#####
	# Tree structure modifier methods
	def add_child(self,n,c,length=None):
		"""
		Add node c to the tree as a child of node n.
		"""
		if c in self.N:
			raise Exception, 'Node %s must be a node not currently in the tree' % str(c)
		
		branch = self.gen_edge(n,c,length)
		
		# update tree data structure
		self.children_lookup[n].append(c)
		self.children_lookup[c] = []
		
		self.parent_lookup[c] = (n,branch)
		
		# update graph data structure
		self.E.add(n,c,branch)

	def remove_child(self,n):
		"""
		Remove the subtree rooted at node n.
		"""
		self.__recur_remove_node(n)
		
	def __recur_remove_node(self,n):
		"""
		Remove node and all its children from the tree.
		"""
		
		if n is self.root_node:
			raise Exception, 'Root cannot be removed'
		
		# remove children
		children = self.children_lookup[n]
		if children is not None:
			for c in self.children_lookup[n]:
				self.__recur_remove_node(c)
				
		# remove self
		del self.children_lookup[n]
		del self.parent_lookup[n]
		
		# update graph data structure
		self.N.remove(n)
	
	def move(self,n,p,length=None):
		"""
		Move node n from where it currently is in the tree to be a child of node p.
		"""
		
		if n == self.root_node:
			raise Exception, 'Cannot move the root'
		
		old_p,b = self.parent_lookup[n]
		
		#####
		# Remove n from its current location
		
		# modify the graph structure
		self.E.remove(b)
		
		# modify tree structure
		del self.parent_lookup[n]
		self.children_lookup[old_p].remove(n)
		
		#####
		# Attach n to its new location
		new_b = self.gen_edge(p,n,length)
		
		# modify graph structure
		self.E.add(p,n,new_b)
		
		# modify tree structure
		self.parent_lookup[n] = (p,new_b)
		self.children_lookup[p].append(n)
		
		return
	
	#####
	# Data structure interrogation methods
	def is_rooted(self):
		"""
		All RTrees and RTree-like objects must return true.  This is the most basic defining
		characteristc of a rooted tree in PyWeaver.
		"""
		return True
	
	def is_leaf(self,node):
		"""
		Return true if the node specified is a leaf.
		"""
		return len(self.children_lookup[node]) == 0
		
	def leaves_iter(self):
		"""
		Return an iterator over the leaves in this tree.
		"""
		for n in self.N:
			if self.is_leaf(n):
				yield n
				
	def leaves(self):
		"""
		Return a list of the leaves in this tree.
		"""
		lvs = []
		for n in self.N:
			if self.is_leaf(n):
				lvs.append(n)
	
		return lvs
	
	def root(self):
		return self.root_node
	
	def parent(self,n):
		"""
		Return the parent of node n.
		"""
		return self.parent_lookup[n][0]
	
	def distance(self,n):
		"""
		Return the distance between this node and its parent
		"""
		if n == self.root_node:
			raise Exception, 'The root does not have a parent distance'
		
		return self.get_edge_length(self.parent_lookup[n][1])
		
	def children(self,n):
		"""
		Return a list of the children of n.
		"""
		cl = self.children_lookup[n]
		
		if cl is None:
			return []
		else:
			return list(cl)
	
	def children_iter(self,n):
		"""
		Return an iterator over the children of n.
		"""
		cl = self.children_lookup[n]
		
		if cl is None:
			return
		else:
			for c in cl:
				yield c
	
	#####
	# Height computing methods
	def __recur_height(self,node,ignore_branch_len):
		clen = max([0.] + [self.__recur_height(c,ignore_branch_len) for c in self.children_iter(node)])
		
		if ignore_branch_len:
			return 1. + clen
		else:
			return self.distance(node) + clen
			
	def height(self,ignore_branch_len=False):
		"""
		Compute the maximum height of this tree.  If ignore_branch_len is True, then each branch length
		is assumed to be 1.
		"""
		return max([self.__recur_height(c,False) for c in self.children_iter(self.root_node)])
	
	#####
	# Newick rendering....
	def newick(self,node2str_fxn=lambda n: str(n)):
		"""
		Return a newick string rendering this tree.
		
		By default the name of a node, n, in the string is the string representation of that 
		node returned by str(n).  This can be overridden by specifying a node2str_fxn that takes
		a node and returns its string representation.
		"""
		return self.__to_newick(self.root_node, node2str_fxn)
		
	def __to_newick(self,n,node2str_fxn):
		"""
		This method does all the work of converting a node (and its children) into a newick string.
		"""
		nstr = ''
		
		#####
		# render children if necessary
		cl = self.children_lookup[n]
		
		if len(cl) > 0:
			nstr += '(' + ', '.join([self.__to_newick(c,node2str_fxn) for c in cl]) + ')'
		
		#####
		# tack on the node's info
		if n != self.root_node:
			p,b = self.parent_lookup[n]
		
			len_str = ''
			if b is not None:
				length = self.get_edge_length(b)
				if length is not None:
					len_str = ':' + str(length)
		
			nstr += node2str_fxn(n) + len_str
		else:
			nstr += node2str_fxn(n) + ';'
		
		return nstr
		
def main():
	pass

if __name__ == '__main__':
	main()

