"""
This module provides various visualization functions that use PyLab (matplotlib).
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

import types

import pylab
from numpy import *

import colors
from pyweaver.cluster.tree import *
import pyweaver.util.distances as distances

__all__ = ['heatmap','Heatmap','SINGLE','COMPLETE']

class Heatmap:
	"""
	This class provides a data structure for containing details about a Heatmap that has been constructed in
	a figure.
	"""
	def __init__(self,X,Y,num_rows,row_labels,axes):
		self.X = X
		self.Y = Y
		self.num_rows = num_rows
		self.row_labels = row_labels
		self.axes = axes

def __recur_apply_ordering(T,node,next_num,ordering,tree_height,tree_scale):
	
	if node == T.root():
		node.x = tree_height
	elif T.is_leaf(node):
		node.x = 0.
	else:
		node.x = T.parent(node).x - (tree_scale * T.distance(node))
		
	if T.is_leaf(node):
		ordering[node.item_idx] = next_num
		node.x = 0.
		node.y = next_num
		return next_num + 1
	else:
		# order children children
		for child in T.children_iter(node):
			next_num = __recur_apply_ordering(T,child,next_num,ordering,tree_height,tree_scale)	
			
		cy = [c.y for c in T.children_iter(node)]
		first = float(min(cy))
		last = float(max(cy))
		
		node.y = (float(last - first) / 2.) + float(first)
		
		return next_num

def __apply_leaf_ordering(T,ordering,tree_height,tree_scale):
	__recur_apply_ordering(T,T.root(),0,ordering,tree_height,tree_scale)
	
def __recur_draw_subtree(T,node,x_offset):
	# draw a line to the parent level
	if node != T.root():
		parent = T.parent(node)
		pylab.plot([node.x + x_offset, parent.x + x_offset], [node.y, node.y],'k')

	if not T.is_leaf(node):
		cpos = [c.y for c in T.children_iter(node)]
		pylab.plot([node.x + x_offset, node.x + x_offset],[min(cpos),max(cpos)],'k')

		for c in T.children_iter(node):
			__recur_draw_subtree(T,c,x_offset)

def __draw_cluster_tree(T,x_offset):
	__recur_draw_subtree(T,T.root(),x_offset)
		
def heatmap(data,**kwargs):
	"""
	This function constructs a heatmap in the current figure/axes based on data, which is an MxN matrix (or a list of M N-entry lists).
	Only the data to render as a heatmap is required.  However, many other options exist:
	
		- pos_color is the color taken by the most positive entry.  All other positive entries are scaled between this color and
		black.  Default is 'green'.
		
		- neg_color is the color taken by the most negative entry.  All other negative entries are scaled between this color and
		black.  Default is 'red'.
		
		- zero_color is the color assigned to values that are zero.  All values scale between this color and the positive or negative
		color, depending on the sign of the value.
		
		- row_labels is a list of strings providing names for each row.  If this list is shorter than the number of rows, the list
		will be repeated.  Default is []
		
		- row_label_rotation is the orientation of the row labels.  Default is 'horizontal'.  Other valid value is 'vertical'.
		
		- col_labels is a list of strings providing names for each column.  If this list is shorter than the number of columns, the
		list will be repeated.  Default is []
		
		- col_label_rotation is the orientation of the column labels.  Default is 'vertical'.  Other valid value is 'horizontal'.
		
		- before/after are two parameters that specify heatmaps already drawn in the present figure that this heatmap should be drawn
		after/before.  The heatmaps must agree in # of rows but can differ in number of columns and in any other respect.  Note that
		a new heatmap can ONLY be safely added as the last or first heatmap - it cannot be inserted between two heatmaps.
		
		- cluster_rows indicates how the rows of the heatmap should be hierarchically clustered.  Default is None indicating no
		clustering.  Other values are the kinds of hierarchical clustering distance methods: SINGLE or COMPLETE.  
		See pyweaver.cluster.tree.hcluster for more information on hierarchical clustering.
		
		- row_cluster_tree_width is the width that the row cluster tree should have when drawn.  Default is 1/6th the width of
		the heatmap.
		
		- row_dist_fxn is a function that accepts two row indices and returns a non-negative number indicating the distance
		between the two rows.  By default, this value is the RMSD of the two rows.
		
	In general, the heatmap completely takes over the axes into which it is drawn.
		
	A tuple of values are returned by this function: (H, C, [I, RT]) where
	
		- H is a heatmap object.  This is primarily used for plotting multiple heatmaps in the same figure.
	
		- C is an MxNx3 matrix where C[i,j,:] are the RGB color components of the (i,j) cell in the heatmap.
		
		- I maps row indices from data to the row indices of C where I[x] = the row in C corresponding to row x in data.  This is
		included only if the rows have been clustered.
		
		- RT is the tree over the original data rows, included only if the rows have been clustered.  Leaves of this tree have
		the item_idx field set to the data row they correspond to.
	"""
	
	## TODO: Add before/after
	
	#####
	# Setup defaults
	pos_color = colors.cword2rgb('green')
	neg_color = colors.cword2rgb('red')
	zero_color = colors.cword2rgb('black')
	
	row_labels = []
	row_label_rotation = 'horizontal'
	col_labels = []
	col_label_rotation = 'vertical'
	
	# this is the default origin.  It will be updated in order to accomodate any before/after conditions.
	origin = (-0.5,-0.5)
	before_hm = None
	after_hm = None
	
	cluster_rows = None
	row_cluster_tree_width = None
	row_dist_fxn = lambda x,y: distances.rmsd(data[x],data[y])
	
	#####
	# Read all keyword argments
	#
	# TODO: check the validity of each parameters
	for k,v in kwargs.items():
		if k == 'pos_color':
			pos_color = v
			
			if type(pos_color) == types.StringType:
				pos_color = colors.cword2rgb(pos_color)
			elif len(pos_color) != 3 or len(pos_color) != 4:
				raise Exception, 'pos_color value must resolve to an RGB or RBGA value'
		elif k == 'neg_color':
			neg_color = v
			
			if type(neg_color) == types.StringType:
				neg_color = colors.cword2rgb(neg_color)
			elif len(neg_color) != 3 or len(neg_color) != 4:
				raise Exception, 'neg_color value must resolve to an RGB or RBGA value'
		elif k == 'zero_color':
			zero_color = v
			
			if type(zero_color) == types.StringType:
				zero_color = colors.cword2rgb(zero_color)
			elif len(zero_color) != 3 or len(zero_color) != 4:
				raise Exception, 'neg_color value must resolve to an RGB or RBGA value'
		elif k == 'row_labels':
			row_labels = v
		elif k == 'row_label_rotation':
			row_label_rotation = v
		elif k == 'col_labels':
			col_labels = v
		elif k == 'col_label_rotation':
			col_label_rotation = v
		elif k == 'before':
			before_hm = v
		elif k == 'after':
			after_hm = v
		elif k == 'cluster_rows':
			cluster_rows = v
			
			if type(cluster_rows) != types.StringType:
				raise Exception, 'cluster_rows value must be a string'
		elif k == 'row_dist_fxn':
			row_dist_fxn = v
			
			if type(row_dist_fxn) != types.FunctionType:
				raise Exception, 'row_dist_fxn value must be a function'
		else:
			raise Exception, 'Keyword %s is unknown' % k

	#####
	# Sanity check on any before/after conditions requested...
	if before_hm is not None and after_hm is not None:
		raise Exception, 'A heatmap can be created before or after an existing heatmap, not both'
		
	if before_hm is not None:
		if pylab.gca() != before_hm.axes:
			raise Exception, 'A heatmap cannot be added before an existing heatmap that does not share the same axes'
	elif after_hm is not None:
		if pylab.gca() != after_hm.axes:
			raise Exception, 'A heatmap cannot be added after an existing heatmap that does not share the same axes'

	#####
	# cluster rows if appropriate
	I = None
	RT = None
	if cluster_rows:
		# compute distances
		dist = zeros((len(data),len(data)))
		for i in range(len(data)):
			for j in range(i+1,len(data)):
				dist[i,j] = dist[j,i] = row_dist_fxn(i,j)
	
		# cluster...
		RT = hcluster(dist,distance_fxn=cluster_rows)
		
		# apply clustering to row ordering
		if row_cluster_tree_width is None:
			row_cluster_tree_width = float(len(data[0])) / 6.
		
		I = [0] * len(data)
		__apply_leaf_ordering(RT,I,row_cluster_tree_width,(row_cluster_tree_width / RT.height()))

		data_p = [(i,d) for i,d in zip(range(len(data)),data)]
		data_p.sort(cmp=lambda x,y: cmp(I[x[0]],I[y[0]]))
		data = [x[1] for x in data_p]
		
	#####
	# Compute heatmap colors
	x = len(data)
	y = len(data[0])
	C = zeros((x,y,3))
	
	Dmax = max([max(line) for line in data])
	Dmin = min([min(line) for line in data])
	
	for i in range(x):
		for j in range(y):
			if data[i][j] >= 0:
				# C[i,j,0] = (data[i][j] / Dmax) * pos_color[0]
				# C[i,j,1] = (data[i][j] / Dmax) * pos_color[1]
				# C[i,j,2] = (data[i][j] / Dmax) * pos_color[2]
				C[i,j,0] = (data[i][j] / Dmax) * (pos_color[0] - zero_color[0]) + zero_color[0]
				C[i,j,1] = (data[i][j] / Dmax) * (pos_color[1] - zero_color[1]) + zero_color[1]
				C[i,j,2] = (data[i][j] / Dmax) * (pos_color[2] - zero_color[2]) + zero_color[2]
			else:
				# C[i,j,0] = (data[i][j] / Dmin) * neg_color[0]
				# C[i,j,1] = (data[i][j] / Dmin) * neg_color[1]
				# C[i,j,2] = (data[i][j] / Dmin) * neg_color[2]
				C[i,j,0] = (data[i][j] / Dmin) * (neg_color[0] - zero_color[0]) + zero_color[0]
				C[i,j,1] = (data[i][j] / Dmin) * (neg_color[1] - zero_color[1]) + zero_color[1]
				C[i,j,2] = (data[i][j] / Dmin) * (neg_color[2] - zero_color[2]) + zero_color[2]
		
	#####
	# Plot heatmap
	
	# determine the origin and dimensions of this heatmap
	axis_lims = list(pylab.axis())
	
	heatmap_height = x
	
	grid_width = y
	heatmap_width = y
	
	row_tree_margin = 0
	
	if cluster_rows:
		row_tree_margin = row_cluster_tree_width + 0.5
		heatmap_width += row_tree_margin
	
	if before_hm:
		origin = (before_hm.X[0] - heatmap_width, before_hm.Y[0])
		axis_lims[0] -= heatmap_width
	elif after_hm:
		origin = (after_hm.X[1], after_hm.Y[0])
		axis_lims[1] += heatmap_width
	else:
		axis_lims[0] = origin[0]
		axis_lims[1] = origin[0] + heatmap_width
		axis_lims[2] = origin[1]
		axis_lims[3] = origin[1] + heatmap_height

	start_x = origin[0]
	end_x = origin[0] + y
	start_y = origin[1]
	end_y = origin[1] + x
	
	# draw the image
	pylab.hold(True)
	pylab.imshow(C,interpolation='nearest',extent=(start_x,end_x,start_y,end_y),origin='lower')

	# set the column labels if necessary
	# TODO: Make this labeling more friendly with current labelings on the axis
	if len(col_labels) > 0:
		ticks = pylab.gca().get_xticks()
		tick_labels = [lbl.get_text() for lbl in pylab.gca().get_xticklabels()]
		
		our_ticks = range(int(origin[0] + 0.5),int(origin[0] + y + 0.5))
		
		if before_hm:
			pylab.gca().set_xticks(our_ticks + ticks)
			pylab.gca().set_xticklabels(col_labels + tick_labels,size=8,rotation=col_label_rotation)
		elif after_hm:
			pylab.gca().set_xticks(ticks + our_ticks)
			pylab.gca().set_xticklabels(tick_labels + col_labels,size=8,rotation=col_label_rotation)
		else:
			pylab.gca().set_xticks(our_ticks)
			pylab.gca().set_xticklabels(col_labels,size=8,rotation=col_label_rotation)
	elif not before_hm and not after_hm:
		pylab.gca().set_xticks([])
		
	if len(row_labels) > 0:
		ticks = pylab.gca().get_yticks()
		tick_labels = [lbl.get_text() for lbl in pylab.gca().get_yticklabels()]

		our_ticks = range(int(origin[1] + 0.5),int(origin[1] + x + 0.5))

		if before_hm:
			pylab.gca().set_yticks(our_ticks + ticks)
			pylab.gca().set_yticklabels(row_labels + tick_labels,size=8,rotation=row_label_rotation)
		elif after_hm:
			pylab.gca().set_yticks(ticks + our_ticks)
			pylab.gca().set_yticklabels(tick_labels + row_labels,size=8,rotation=row_label_rotation)
		else:
			pylab.gca().set_yticks(our_ticks)
			pylab.gca().set_yticklabels(row_labels,size=8,rotation=row_label_rotation)
	elif not before_hm and not after_hm:
		pylab.gca().set_yticks([])

	# draw the row tree if requested
	if cluster_rows:
		__draw_cluster_tree(RT,end_x)
	
	#####
	# reformat and reposition the axes to make them look good
	dy = axis_lims[3] - axis_lims[2]
	dx = axis_lims[1] - axis_lims[0]
	
	if dy > dx:
		pylab.gca().set_aspect(float(dx) / float(dy),'box','C')
	else:
		pylab.gca().set_aspect(float(dy) / float(dx),'box','C')
		
	# resize the axis limits to contain this heatmap and any others in the figure.
	pylab.axis(axis_lims)
				
	pylab.gca().set_position([0.05,0.05,0.9,0.9],'both')
	
	pylab.hold(False)
	
	# done!
	H = Heatmap((start_x,end_x),(start_y,end_y),x,list(row_labels),pylab.gca())
	if cluster_rows:
		return (H, C, I, RT)
	else:
		return (H, C)

def main():
	pass


if __name__ == '__main__':
	main()

