"""
This module provides classes and functions for visualizing PyWeaver objects using the Tk windowing
environment.
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

import types
import math
import thread
import time
from random import *

from Tkinter import *

import pyweaver.constants as constants
import pyweaver.graph as g_mod

import tk_geometry as tkg
import colors

__all__ = ['GraphCanvas','GraphFrame','GraphViewer','view','test_GraphCanvas']

####
# Some constants for rendering

# Indexes into the node2had_render_hints hashtable in GraphCanvas
SELECT_EPSILON = 1
SELECT_WIDTH = 3

NODE_X_IDX = 0
NODE_Y_IDX = 1
NODE_COLOR_IDX = 2
NODE_LABEL_IDX = 3
NODE_SIZE_IDX = 4

EDGE_COLOR_IDX = 0
EDGE_WIDTH_IDX = 1
EDGE_WAYPOINTS_IDX = 2
EDGE_HEAD_IDX = 3

ARROW_SHAPE = (8,10,3)
ARROW_SHAPE_STR = '8 10 3'
PLUNGER_SHAPE = (3,1,8)
PLUNGER_SHAPE_STR = '3 1 8'

DEFAULT_NODE_COLOR = "blue"
DEFAULT_NODE_RADIUS = 5

DEFAULT_EDGE_COLOR = "black"
DEFAULT_EDGE_WIDTH = 1

DEFAULT_WIN_SIZE = (200,200)

WAYPOINT_RADIUS = 2
TEXT_Y_BUFFER = 5
MAX_WAYPOINT_INSERTION_DIST = 5
	
# Canvas event callaback function
def remove_waypoint(event):
	"""
	Handles a UI event that should remove the waypoint at the location of the event.
	"""	
	gc = event.widget
	x = gc.canvasx(event.x)
	y = gc.canvasy(event.y)
	
	picked_item = gc.find_closest(x,y)[0]

	if picked_item not in gc.waypoint_info:
		return
		
	edge = gc.waypoint_info[picked_item]
	src, tgt, line_id_list, waypoint_list = gc.edge2graphics[edge]
	
	widx = waypoint_list.index(picked_item)
	line_id1 = line_id_list[widx]
	line_id2 = line_id_list[widx+1]
	
	gonner_line_id = None
	
	if line_id2 == line_id_list[-1]:
		gonner_line_id = line_id1
		
		tgt_coords = gc.coords(gc.node2graphics[tgt][0])
		if line_id1 == line_id_list[0]:
			src_coords = gc.coords(gc.node2graphics[src][0])
			gc.coords(line_id2,tkg.compute_line_coords_from_oval_info(src_coords,tgt_coords,gc.node2size[src],gc.node2size[tgt]))
		else:		
			src_coords = gc.coords(line_id1)
			tgt_coords = tkg.update_tgt_on_circle((src_coords[0],src_coords[1]),(tgt_coords[0]+gc.node2size[tgt],tgt_coords[1]+gc.node2size[tgt]),gc.node2size[tgt])
			gc.coords(line_id2,(src_coords[0],src_coords[1],tgt_coords[0],tgt_coords[1]))
	else:
		gonner_line_id = line_id2
		
		tgt_coords = gc.coords(line_id2)
		if line_id1 == line_id_list[0]:
			src_coords = gc.coords(gc.node2graphics[src][0])
			src_coords = tkg.update_tgt_on_circle((tgt_coords[0],tgt_coords[1]),(src_coords[0]+gc.node2size[src],src_coords[1]+gc.node2size[src]),gc.node2size[src])
			gc.coords(line_id1,(src_coords[0],src_coords[1],tgt_coords[2],tgt_coords[3]))
		else:
			src_coords = gc.coords(line_id1)
			gc.coords(line_id1,(src_coords[0],src_coords[1],tgt_coords[2],tgt_coords[3]))
		
	# delete waypoint info
	gc.delete(picked_item)
	del gc.waypoint_info[picked_item]
	waypoint_list.remove(picked_item)
	
	# delete line info
	gc.delete(gonner_line_id)
	line_id_list.remove(gonner_line_id)
	del gc.edge_line_info[gonner_line_id]
	

def pick_node(event, picked_item):
	"""
	pick_component method delegates the picking of nodes to this method
	"""
	gc = event.widget
	x = gc.canvasx(event.x)
	y = gc.canvasy(event.y)
	
	selected_node = gc.node_oval_info[picked_item][0]
	
	pos = gc.coords(picked_item)
	dx = ((pos[0] + pos[2]) / 2) - x
	dy = ((pos[1] + pos[3]) / 2) - y

	gc.picked_item = picked_item
	gc.picked_node = selected_node
	gc.picked_dx = dx
	gc.picked_dy = dy


def pick_line(event,selected_line):
	"""
	pick_component method delegates the picking of nodes to this method
	"""
	gc = event.widget
	x = gc.canvasx(event.x)
	y = gc.canvasy(event.y)

	# Get the closest point on the line to this point
	line_coords = gc.coords(selected_line)
	dist = tkg.compute_pt2line_distance((x,y),line_coords)
	
	# if the user clicked close enough to the line, insert a waypoint
	if dist > MAX_WAYPOINT_INSERTION_DIST:
		return
	
	# get the line properties
	line_width = gc.itemcget(selected_line,'width')
	line_color = gc.itemcget(selected_line,'fill')
	
	# get the edge that will be extended
	edge = gc.edge_line_info[selected_line]
	einfo = gc.edge2graphics[edge]
	line_id_list = einfo[2]
	waypoint_list = einfo[3]
	
	idx = line_id_list.index(selected_line)
	line_id1 = None
	line_id2 = None
	
	waypoint_insertion_idx = None
	
	if idx == (len(line_id_list) - 1):
		line_id2 = selected_line
		line2_coords = list(gc.coords(line_id2))
		gc.coords(line_id2,(x,y,line2_coords[2],line2_coords[3]))
		
		line_id1 = gc.create_line((line2_coords[0],line2_coords[1],x,y), width=line_width, fill=line_color)
	
		# Add the new line into the edge lines
		line_id_list.insert(idx, line_id1)
		gc.edge_line_info[line_id1] = edge
		
		waypoint_insertion_idx = len(waypoint_list)
	else:
		line_id1 = selected_line
		line1_coords = gc.coords(line_id1)
		gc.coords(line_id1,(line1_coords[0],line1_coords[1],x,y))
		
		line_id2 = gc.create_line((x,y,line1_coords[2],line1_coords[3]), width=line_width, fill=line_color)
	
		# Add the new line into the edge lines
		line_id_list.insert(idx+1, line_id2)
		gc.edge_line_info[line_id2] = edge
		
		waypoint_insertion_idx = idx
		
	# draw the waypoint
	oval_id = gc.create_oval((x-WAYPOINT_RADIUS,y-WAYPOINT_RADIUS,x+WAYPOINT_RADIUS,y+WAYPOINT_RADIUS),
							 width=line_width, fill=line_color)
	
	waypoint_list.insert(waypoint_insertion_idx,oval_id)
	
	gc.waypoint_info[oval_id] = edge
	
	# Now that the waypoint has been created, pick it
	pick_waypoint(event, oval_id)


def pick_waypoint(event, picked_item):
	"""
	pick_component method delegates the picking of waypoints to this method
	"""
	gc = event.widget
	gc.picked_item = picked_item
	
	edge = gc.waypoint_info[gc.picked_item]
	src, tgt, line_id_list, waypoint_list = gc.edge2graphics[edge]
	
	widx = waypoint_list.index(picked_item)
	line_id1 = line_id_list[widx]
	line_id2 = line_id_list[widx+1]
	
	line_coords = gc.coords(line_id1)

	src_coords = None
	tgt_coords = None

	if line_id1 == line_id_list[0]:
		coords = gc.coords(gc.node2graphics[src][0])
		# Position of the target is the center of the circle
		src_coords = (coords[0] + gc.node2size[src], coords[1] + gc.node2size[src])
		
	line_coords = gc.coords(line_id2)
		
	if line_id2 == line_id_list[-1]:
		coords = gc.coords(gc.node2graphics[tgt][0])
		# Position of the target is the center of the circle
		tgt_coords = (coords[0] + gc.node2size[tgt], coords[1] + gc.node2size[tgt])	
		
	gc.picked_waypoint_info = (line_id1, line_id2, src_coords, tgt_coords, src, tgt)
		

def pick_component(event):
	"""
	Handles a UI event that should pick a component in the GraphCanvas"
	"""
	gc = event.widget
	x = gc.canvasx(event.x)
	y = gc.canvasy(event.y)
	
	picked_item = gc.find_closest(x,y)[0]

	# what item did we pick?
	if picked_item in gc.node_oval_info:
		pick_node(event, picked_item)
	if picked_item in gc.edge_line_info:
		pick_line(event, picked_item)
	if picked_item in gc.waypoint_info:
		pick_waypoint(event, picked_item)
		

def handle_btn1up(event):
	gc = event.widget
	gc.picked_item = None
	gc.picked_node = None
	gc.picked_waypoint_info = None
	
def drag_on_move(event):
	"""
	Handle graph layout editing mouse movements.  These will drag items if they are
	selected.
	"""
	gc = event.widget
	x = gc.canvasx(event.x)
	y = gc.canvasy(event.y)
		
	if gc.picked_waypoint_info:
		# relocate the waypoint
		oval_coords = gc.coords(gc.picked_item)
		gc.coords(gc.picked_item,(x-WAYPOINT_RADIUS,y-WAYPOINT_RADIUS,x+WAYPOINT_RADIUS,y+WAYPOINT_RADIUS))
		
		###
		# update the waypoints lines
		line_id1, line_id2, src_coords, tgt_coords, src_node, tgt_node = gc.picked_waypoint_info

		# Move line 1
		line_coords = gc.coords(line_id1)
		
		if src_coords is not None:
			src = tkg.update_tgt_on_circle((x,y),src_coords,gc.node2size[src_node])
			gc.coords(line_id1,(src[0],src[1],x,y))
		else:
			gc.coords(line_id1,(line_coords[0],line_coords[1],x,y))
		
		# Move line 2
		line_coords = gc.coords(line_id2)
		
		if tgt_coords is not None:
			tgt = tkg.update_tgt_on_circle((x,y),tgt_coords,gc.node2size[tgt_node])
			gc.coords(line_id2,(x,y,tgt[0],tgt[1]))
		else:
			gc.coords(line_id2,(x,y,line_coords[2],line_coords[3]))
		
	elif gc.picked_node:
		gc.move_node(gc.picked_node, x + gc.picked_dx, y + gc.picked_dy)
							
def select_component(event):
	"""
	Select the component nearest to the click event
	"""
	gc = event.widget
	x = gc.canvasx(event.x)
	y = gc.canvasy(event.y)
	
	clicked_items = gc.find_overlapping(x - SELECT_EPSILON, y - SELECT_EPSILON, x + SELECT_EPSILON, y + SELECT_EPSILON)
	
	clicked_item = None
	if len(clicked_items) > 0:
		clicked_item = clicked_items[0]

	# what item did we select?
	if clicked_item in gc.node_oval_info:
		gc.set_selected(gc.node_oval_info[clicked_item][0])
	elif clicked_item in gc.edge_line_info:
		gc.set_selected(gc.edge_line_info[clicked_item])
	elif clicked_item in gc.waypoint_info:
		gc.set_selected(gc.waypoint_info[clicked_item])
	else:
		gc.set_selected(None)

def multiselect_component(event):
	"""
	When control is selected, multiple nodes can be selected.  If a selected node is clicked, then it is deselected.
	"""
	gc = event.widget
	x = gc.canvasx(event.x)
	y = gc.canvasy(event.y)
	
	clicked_items = gc.find_overlapping(x - SELECT_EPSILON, y - SELECT_EPSILON, x + SELECT_EPSILON, y + SELECT_EPSILON)
	clicked_item = None
	
	if len(clicked_items) > 0:
		clicked_item = clicked_items[0]

	# what item did we select?
	if clicked_item in gc.node_oval_info:
		gc.toggle_selected(gc.node_oval_info[clicked_item][0])
	elif clicked_item in gc.edge_line_info:
		gc.toggle_selected(gc.edge_line_info[clicked_item])
	elif clicked_item in gc.waypoint_info:
		gc.toggle_selected(gc.waypoint_info[clicked_item])
	else:
		pass

	
# GraphCanvas class
class GraphCanvas(Canvas):
	"""
	A Tkinter Canvas that displays and allows the user to interact with a graph.  In this version
	self-loops are not rendered.
	"""
	def __init__(self,root,graph,default_node_radius=None):
		Canvas.__init__(self,root)
		
		self.graph = graph
		
		self.default_ncolor = DEFAULT_NODE_COLOR
		self.default_ecolor = DEFAULT_EDGE_COLOR
		
		# handle input parameters
		if default_node_radius:
			self.default_node_size = float(default_node_radius)
		else:
			self.default_node_size = float(DEFAULT_NODE_RADIUS)
		
		self._build_ui()
		
	# Canvas methods
	def _build_ui(self):
		"""
		This method is an internal method for setting up the visualization window
		"""
		
		# determine the max size of this canvas
		max_x = 0
		max_y = 0
		
		# Configure the canvas
		self.bind('<Button-1>',select_component)
		self.bind('<Control-Button-1>',multiselect_component)
		self.bind("<Shift-Button-1>",pick_component)
		self.bind("<Shift-Button-2>",remove_waypoint)
		self.bind("<B1-Motion>",drag_on_move)
		self.bind("<ButtonRelease-1>",handle_btn1up)
		
		# layout all the nodes
		self.node2size = {}
		self.node2is_default = {}
		self.node2graphics = {}
		self.node_oval_info = {}
		
		for node in self.graph.N:
			
			# Apply any rendering hints that this node has
			has_x = False
			has_y = False
			has_color = False
			has_label = False
			has_size = False
			
			if hasattr(node,'x'):
				x = node.x
				has_x = True
			else:
				x = randint(0,DEFAULT_WIN_SIZE[0]-self.default_node_size)
				
			if hasattr(node,'y'):
				y = node.y
				has_y = True
			else:
				y = randint(0,DEFAULT_WIN_SIZE[1]-self.default_node_size)
			
			if hasattr(node,'color'):
				color = colors.color2tkrgb(node.color)
				has_color = True
			else:
				color = self.default_ncolor
			
			if hasattr(node,'label'):
				label = node.label
				has_label = True
			elif hasattr(node,'__str__'):
				label = node.__str__()
				has_label = True
			else:
				label = ''
				
			if hasattr(node,'size'):
				size = node.size
				has_size = True
			else:
				size = self.default_node_size
			
			# store knowledge of render hints for later
			self.node2is_default[node] = [not has_x,not has_y,not has_color,not has_label,not has_size]
			
			oval_id = self.create_oval((x - size, y - size, x + size, y + size), fill=color)
			txt_id = self.create_text(x + size, y - size - TEXT_Y_BUFFER, text=label, anchor=NW)
			self.node2size[node] = size
			self.node2graphics[node] = (oval_id,txt_id)
			self.node_oval_info[oval_id] = (node,txt_id)

			if (x + size) > max_x:
				max_x = x + size
				
			if (y + size) > max_y:
				max_y = y + size

		# draw edges
		self.edge2is_default = {}
		self.edge_width = {}
		self.edge2graphics = {}
		self.edge_line_info = {}
		self.waypoint_info = {}
		
		is_directed = self.graph.is_directed()
			 
		for edge in self.graph.E:
		
			had_color = False
			had_width = False
			had_waypoints = False
			has_head = False

			if is_directed:
				src = self.graph.source(edge)
				tgt = self.graph.target(edge)
			else:
				ep = self.graph.endpoints(edge)
				src = ep[0]
				tgt = ep[1]		
		
			# don't draw self-loops
			if src == tgt:
				continue
		
			# check for head attribute
			if hasattr(edge,'head'):
				had_head = True
		
			# Generate waypoint list
			waypoints = []
			soval = self.coords(self.node2graphics[src][0])
			toval = self.coords(self.node2graphics[tgt][0])
			endpts = tkg.compute_line_coords_from_oval_info(soval, toval, self.node2size[src], self.node2size[tgt])
			
			waypoints.append((endpts[0],endpts[1]))
			
			# check for the waypoints render hint
			if hasattr(edge,"waypoints"):
				had_waypoints = True
				waypoints = waypoints + edge.waypoints
			
			waypoints.append((endpts[2],endpts[3]))
			
			###
			# Create the lines that form the edge
			
			# select the edge color
			line_color = self.default_ecolor
			
			if hasattr(edge,'color'):
				had_color = True
				line_color = colors.color2tkrgb(edge.color)
			
			# set the edge width
			line_width = DEFAULT_EDGE_WIDTH
			
			if hasattr(edge,'width'):
				had_width = True
				line_width = edge.width
			
			self.edge_width[edge] = line_width
			self.edge2is_default[edge] = [not had_color, not had_width, not had_waypoints, not has_head]
			
			# create the individual lines
			line_id_list = []
			waypoint_list = []
			last_line_id = None
			
			for i in range(1,len(waypoints)):
				p1 = waypoints[i-1]
				p2 = waypoints[i]
				line_id = self.create_line((p1[0],p1[1],p2[0],p2[1]), width=line_width, fill=line_color)
				line_id_list.append(line_id)
				
				if len(waypoints) > 2 and i < (len(waypoints) - 1):
					oval_id = self.create_oval((p2[0]-WAYPOINT_RADIUS,p2[1]-WAYPOINT_RADIUS,
												p2[0]+WAYPOINT_RADIUS,p2[1]+WAYPOINT_RADIUS), fill=line_color)
					self.waypoint_info[oval_id] = edge
					waypoint_list.append(oval_id)
				
				last_line_id = line_id
				
				if p1[0] > max_x:
					max_x = p1[0]
				
				if p2[0] > max_x:
					max_x = p2[0]
					
				if p1[1] > max_y:
					max_y = p1[1]
				
				if p2[1] > max_y:
					max_y = p2[1]				
			
			# put an arrowhead on the last line of the edge if appropriate
			# change the arrow to the indicated one if a render hint is provided.
			if is_directed:
				if not hasattr(edge,"head"): # Draw an arrow by default
					self.itemconfig(line_id_list[-1], arrowshape=ARROW_SHAPE, arrow="last")
				elif edge.head == ">": # Explicitly want an arrow
					self.itemconfig(line_id_list[-1], arrowshape=ARROW_SHAPE, arrow="last")
				elif edge.head == "|": # Want a plunger
					self.itemconfig(line_id_list[-1], arrowshape=PLUNGER_SHAPE, arrow="last")
				elif edge.head == '-' or edge.head is None:
					# don't add a head at all
					pass
				
			self.edge2graphics[edge] = (src,tgt,line_id_list,waypoint_list)
			
			for line_id in line_id_list:
				self.edge_line_info[line_id] = edge

		# set the scrollable region
		self.config(scrollregion=(0, 0, max_x, max_y))

		# setup picking system
		self.picked_item = None
		self.picked_node = None
		self.picked_waypoint_info = None
		self.picked_dx = 0
		self.picked_dy = 0
		
		# setup selection system
		self.selected_components = set()
		self.selection_listeners = []

	######
	# Selection system
	def selected(self):
		"""
		Return a set containing the currently selected nodes.  This set is guaranteed to be independent
		of the underlying implementation and, therefore, may be modified.
		"""
		return set(self.selected)
	
	def set_selected(self,component=None):
		"""
		Set the selected components to the specified object.  This automatically deselects all other 
		currently selected components.  If the component to be selected is already selected, then
		no deselection (or selection) is done.
		"""
		# don't do anything if the component is in the selected component list
		if component is not None and component in self.selected_components:
			return
		
		# deselect all currently selected 
		if len(self.selected_components) > 0:
			self._notify_selection_listeners_of_multiremove(self.selected_components)
		
		for c in self.selected_components:
			if c in self.graph.V:
				self.itemconfig(self.node2graphics[c][0],width=1)
			elif self.graph.E.contains(c):
				for lid in self.edge2graphics[c][2]:
					self.itemconfig(lid,width=self.edge_width[c])
			else:
				raise Exception, 'Unknown component type: ' + c.__str__()
				
			
		self.selected_components.clear()
		
		# select this component
		if component is not None:
			if component in self.graph.V:
				self.itemconfig(self.node2graphics[component][0],width=SELECT_WIDTH)
			elif self.graph.E.contains(component):
				for lid in self.edge2graphics[component][2]:
					self.itemconfig(lid,width=SELECT_WIDTH)
	
			self.selected_components.add(component)
			self._notify_selection_listeners_of_add(component)
		
	def toggle_selected(self,component):
		
		# deselect if already selected
		if component in self.selected_components:
			# visually deselect the node
			if compontns in self.graph.V:
				self.itemconfig(self.node2graphics[component][0],width=1)
			elif self.graph.E.contains(component):
				for lid in self.edge2graphics[component][2]:
					self.itemconfig(lid,width=self.edge_width[component])
			else:
				raise Exception, 'Unknown component type: ' + component.__str__()
			
			# the node is no longer selected
			self.selected_components.remove(component)
			self._notify_selection_listeners_of_remove(component)			
		else: # select if not currently selected
			if component in self.graph.V:
				self.itemconfig(self.node2graphics[component][0],width=SELECT_WIDTH)
			elif self.graph.E.contains(component):
				for lid in self.edge2graphics[component][2]:
					self.itemconfig(lid,width=SELECT_WIDTH)
	
			self.selected_components.add(component)
			self._notify_selection_listeners_of_add(component)
			
		
	def _notify_selection_listeners_of_add(self,c):
		for sl in self.selection_listeners:
			if hasattr(sl,'component_selected'):
				sl.component_selected(c)
			elif hasattr(sl,'components_selected'):
				sl.components_selected([c])
	
	def _notify_selection_listeners_of_remove(self,c):
		for sl in self.selection_listeners:
			if hasattr(sl,'component_deselected'):
				sl.component_deselected(c)
			elif hasattr(sl,'components_deselected'):
				sl.components_deselected([c])

	def _notify_selection_listeners_of_multiremove(self,cl):
		for sl in self.selection_listeners:
			if hasattr(sl,'components_deselected'):
				sl.components_deselected(cl)
			elif hasattr(sl,'component_deselected'):
				for c in cl:
					sl.component_deselected(c)

	def add_selection_listener(self,sl):
		self.selection_listeners.append(sl)
		
	def rm_selection_listener(self,sl):
		if sl in self.selection_listeners:
			self.selection_listeners.remove(sl)
			
	###
	# Synchronization functions
	def sync_graph(self,force_update=False):
		"""
		Write the current rendering properties of the graph in the canvas to the underlying graph's 
		render hints.
		
		If force_update is False, then only those render hints that objects have are updated.
		If force_update is True, then if an object does not have a render hint, it will be created
		and then updated.
		"""

		# update all node rendering hints
		for n,g in self.node2graphics.items():
			oval_id, txt_id = g
			
			coords = self.coords(oval_id)
			
			if hasattr(n,'x') or force_update:
				n.x = coords[0] + self.node2size[n]
			if hasattr(n,'y') or force_update:
				n.y = coords[1] + self.node2size[n]
			if hasattr(n,'color') or force_update:
				n.color = self.itemcget(oval_id,'fill')
			if hasattr(n,'label') or force_update:
				n.label = self.itemcget(txt_id,'text')
			if hasattr(n,'size') or force_update:
				cds = self.coords(oval_id)
				n.size = (cds[2] - cds[0]) / 2.
			
		# update all edge rendering hints
		for e,g in self.edge2graphics.items():
			src, tgt, line_id_list, waypoint_list = g
			
			if hasattr(e,'head') or force_update:
				if self.itemcget(line_id_list[-1],'arrow') != 'NONE':
					ashape = self.itemcget(line_id_list[-1],'arrowshape')
					
					if ashape == ARROW_SHAPE_STR:
						e.head = '>'
					elif ashape == PLUNGER_SHAPE_STR:
						e.head = '|'
					else:
						raise Exception, 'Unknown head shape: ' + ashape.__str__()
				else:
					e.head = '-'
			if hasattr(e,'color') or force_update:
				e.color = self.itemcget(line_id_list[0],'fill')
			if hasattr(e,'width') or force_update:
				e.width = self.itemcget(line_id_list[0],'width')
			if hasattr(e,'waypoints') or force_update:
				# write waypoints to the edge
				if len(waypoint_list) > 0:
					e.waypoints = []
					for w in waypoint_list:
						oval_pos = self.coords(w)
						e.waypoints.append((oval_pos[0] + WAYPOINT_RADIUS, oval_pos[1] + WAYPOINT_RADIUS))
				else:
					e.waypoints = []
					
	###
	# Color methods...
	def default_node_color(self,color=None,reapply=True):
		"""
		Return the default color applied to nodes in the graph.
		
		If color is specified then the change is applied to all nodes that were not added with
		a color render hint.  If reapply is False, then the color of nodes already being 
		visualized are not changed.  If reapply is True, then the color of nodes whose color was
		not specified in a render hint or via a call to color(...) will be changed.
		"""
		if color:
			self.default_ncolor = color
		else:
			return self.default_ncolor
			
		# apply the color to all nodes whose color is default
		if reapply:
			for n,v in self.node2is_default.items():
				if v[NODE_COLOR_IDX]:
					self.itemconfig(self.node2graphics[n][0],fill=colors.color2tkrgb(color))
	
	def default_edge_color(self,color=None,reapply=True):
		"""
		Return the default color applied to edges in the graph.
		
		If color is specified then the change is applied to all edges added to the graph after
		the change.  If reapply is False, then any edges currently being visualized are not affected.
		If reapply is True, then any edges whose colors were not specified in a render hint or by a call
		to the color(...) method will have their color changed.
		"""	
		if color:
			self.default_ecolor = color
		else:
			return self.default_ecolor

		# apply the color to all edges whose color is default
		if reapply:
			for e,v in self.edge2is_default.items():
				# apply the color to each line in the edge
				if v[EDGE_COLOR_IDX]:
					for lid in self.edge2graphics[e][2]:
						self.itemconfig(lid,fill=colors.color2tkrgb(color))
						
	###
	# color(...)
	def _get_color(self,component):
		"""
		Helper method for GraphCanvas.colors(...).  This method returns the color of the specified
		component(s)
		"""
		if type(component) is types.ListType:
			color_list = []
			for c in component:
				color_list.append(self._get_color(c))
				
			return color_list
			
		elif component == constants.NODES:
			color_list = []
			for c in self.network.N:
				color_list.append(self._get_color(c))
				
			return color_list
			
		elif component == constants.EDGES:
			color_list = []
			for c in self.network.E:
				color_list.append(self._get_color(c))
				
			return color_list
			
		elif component == constants.ALL:
			color_list = self._get_color(constants.NODES)
			color_list += self._get_color(constants.EDGES)
			
			return color_list
			
		elif component in self.node2graphics:
			oval_id, txt_id = self.node2graphics[component]
			return colors.tkrgb2color(self.itemcget(oval_id,'fill'))
			
		elif component in self.edge2graphics:
			(src,tgt,line_id_list,waypoint_list) = self.edge2graphics[component]
			return colors.tkrgb2color(self.itemconfigure(line_id_list[0],'fill'))
			
		else:
			raise Exception, 'Component ' + str(c) + ' not found in graph'
		
	def color(self,component,color=None):
		"""
		If color is specified, set the color of the component(s) to the color indicated.
		
		If color is not specified, return the color(s) of the component(s).
		
		In both uses of this method, NODES, EDGES, and ALL can be used as components.
		"""
		if color is None:
			return self._get_color(component)
		
		# convert to Tk color string
		rgb_str = colors.color2tkrgb(color)
		
		if component == constants.NODES:
			# Set the color for all nodes
			for n,g in self.node2graphics.items():
				self.node2is_default[n][NODE_COLOR_IDX] = False
				
				oval_id,txt_id = g
				self.itemconfigure(oval_id,fill=rgb_str)

		elif component == constants.EDGES:
			# set the color for all edges
			for e,g in self.edge2graphics.items():
				self.edge2is_default[e][EDGE_COLOR_IDX] = False
				
				line_id_list = g[2]
				
				for lid in line_id_list:
					self.itemconfigure(lid,fill=rgb_str)
			
		elif component == constants.ALL:
			# set the color for all edges and nodes
			self.color(constants.NODES,color)
			self.color(constants.EDGES,color)
			
		elif type(component) is types.ListType:
			# set the color for all components in the list
			for c in component:
				self.color(c,color)
			
		elif component in self.node2graphics:
			# set the color for just this node
			self.node2is_default[component][NODE_COLOR_IDX] = False
			
			oval_id, txt_id = self.node2graphics[component]
			
			self.itemconfigure(oval_id,fill=rgb_str)
		
		elif component in self.edge2graphics:
			# set the color for this edge
			self.edge2is_default[component][EDGE_COLOR_IDX] = False			
			
			(src,tgt,line_id_list,waypoint_list) = self.edge2graphics[component]
			
			for lid in line_id_list:
				self.itemconfigure(lid,fill=rgb_str)

	###
	# width(...)
	def _get_width(self,e):
		"""
		A helper method for width(...).  Returns the width of edge(s) e.
		"""
	
		if e == constants.EDGES or e == constants.ALL:
			widths = []
			
			for g in self.edge2graphics.values():
				line_id_list = g[2]
				
				widths.append(self.itemcget(line_id_list[0],'width'))
					
			return widths
			
		elif type(e) is types.ListType:
			widths = []
			
			for edge in e:
				widths.append(self.width(edge))
		
			return widths
		
		elif e in self.edge2graphics:
			g = self.edge2graphics[e]
			
			return self.itemcget(g[2][0],'width')
		else:
			raise Exception, 'Edge(s) ' + str(e) + ' not found in graph' 
	
	def width(self,e,w=None):
		"""
		Set the width of the specified edge(s).  If no width is specified, then the width of the 
		specified edges is returned.  When e is EDGES or ALL, this method returns the width of all
		edges.
		"""
		if w is None:
			return self._get_width(e)
			
		if e == constants.EDGES or e == constants.ALL:
			# set the widths of all edges
			widths = []
			for e,g in self.edge2graphics.items():
				self.edge2is_default[e][EDGE_WIDTH_IDX] = False
				self.edge_width[e] = w
				
				line_id_list = g[2]
				
				for lid in line_id_list:
					self.itemconfigure(lid,width=w)
					
		elif type(e) is types.ListType:
			# set the width of all specified edges
			for edge in e:
				self.width(edge,w)
		
		elif e in self.edge2graphics:
			self.edge2is_default[e][EDGE_WIDTH_IDX] = False
			self.edge_width[e] = w
			g = self.edge2graphics[e]
			
			for lid in g[2]:
				self.itemconfigure(lid,width=w)
		else:
			raise Exception, 'Edge(s) ' + str(e) + ' not found in graph'
		
	###
	# head(...)
	def _get_head(self,e):
		"""
		A helper method for head(...).  Returns the head of edge(s) e.
		"""
	
		if e == constants.EDGES or e == constants.ALL:
			heads = []
			
			for g in self.edge2graphics.values():
				line_id_list = g[2]
				
				has_head = self.itemcget(line_id_list[-1],'arrow')
				head = self.itemcget(line_id_list[-1],'arrowshape')
			
				head_symbol = '-'
				
				if has_head is None:
					# If no head was specified...
					head_symbol = '-'
				elif head == ARROW_SHAPE:
					head_symbol = '>'
				elif head == PLUNGER_SHAPE:
					head_symbol = '|'
				
				heads.append(head_symbol)
					
			return heads
			
		elif type(e) is types.ListType:
			heads = []
			
			for edge in e:
				heads.append(self._get_heads(edge))
		
			return heads
		
		elif e in self.edge2graphics:
			g = self.edge2graphics[e]
			
			has_head = self.itemcget(line_id_list[-1],'arrow')
			head = self.itemcget(line_id_list[-1],'arrowshape')
		
			head_symbol = '-'
			
			if has_head is None:
				# If no head was specified...
				head_symbol = '-'
			elif head == ARROW_SHAPE:
				head_symbol = '>'
			elif head == PLUNGER_SHAPE:
				head_symbol = '|'
			
			return head_symbol
		else:
			raise Exception, 'Edge(s) ' + str(e) + ' not found in graph' 
	
	def head(self,e,head=None):
		"""
		Set the head of the specified edge(s).  If no head is specified, then the head of the 
		specified edges is returned.  When e is EDGES or ALL, this method returns the head of all
		edges.
		"""
		if head is None:
			return self._get_head(e)
		
		# No sense trying to set the head of an edge in an undirected graph
		if not self.graph.is_directed():
			return
		
		if e == constants.EDGES or e == constants.ALL:
			# set the heads of all edges
			widths = []
			for e,g in self.edge2graphics.items():
				self.edge2is_default[e][EDGE_HEAD_IDX] = False
				last_line_id = g[2][-1]
				
				if head == ">": # Explicitly want an arrow
					self.itemconfig(last_line_id, arrowshape=ARROW_SHAPE, arrow="last")
				elif head == "|": # Want a plunger
					self.itemconfig(last_line_id, arrowshape=PLUNGER_SHAPE, arrow="last")
				else:
					# the last segment shouldn't have a head at all
					self.itemconfig(last_line_id, arrow=None)
					
		elif type(e) is types.ListType:
			# set the head of all specified edges
			for edge in e:
				self.head(edge,head)
		
		elif e in self.edge2graphics:
			self.edge2is_default[e][EDGE_HEAD_IDX] = False
			g = self.edge2graphics[e]
			
			lid = g[2][-1]
			
			if head == ">": # Want an arrow
				self.itemconfig(last_line_id, arrowshape=ARROW_SHAPE, arrow="last")
			elif head == "|": # Want a plunger
				self.itemconfig(last_line_id, arrowshape=PLUNGER_SHAPE, arrow="last")
			else:
				# the last segment shouldn't have a head at all
				self.itemconfig(last_line_id, arrow=None)			


		else:
			raise Exception, 'Edge(s) ' + str(e) + ' not found in graph'
		
	###
	# waypoints(...)
	def _get_waypoints(self,e):
		"""
		A helper method for waypoints(...).  Returns the waypoints of edge(s) e.
		"""
	
		if e == constants.EDGES or e == constants.ALL:
			waypoints_list = []
			
			for g in self.edge2graphics.values():
				wlist = g[3]
				
				wp_list = []
				
				for ovalid in wlist:
					coords = self.coords(ovalid)
					wp_list.append( (coords[0] + WAYPOINT_RADIUS, coords[1] + WAYPOINT_RADIUS) )
				
				wapoints_list.append(wp_list)
				
			return waypoints_list
			
		elif type(e) is types.ListType:
			waypoints_list = []
			
			for edge in e:
				g = self.edge2graphics[edge]
				wlist = g[3]
				
				wp_list = []
				
				for ovalid in wlist:
					coords = self.coords(ovalid)
					wp_list.append( (coords[0] + WAYPOINT_RADIUS, coords[1] + WAYPOINT_RADIUS) )
				
				wapoints_list.append(wp_list)
				
			return waypoints_list
		
		elif e in self.edge2graphics:
			g = self.edge2graphics[e]
			wlist = g[3]
			
			wp_list = []
			
			for ovalid in wlist:
				coords = self.coords(ovalid)
				wp_list.append( (coords[0] + WAYPOINT_RADIUS, coords[1] + WAYPOINT_RADIUS) )
				
			return wp_list
		else:
			raise Exception, 'Edge(s) ' + str(e) + ' not found in graph' 
		
	def waypoints(self,e,wps=None):
		"""
		The waypoints of the specified edges are returned.  A single edge, a list of edges, or EDGES or ALL are permitted 
		and this method returns the waypoints of the specified set of edges.
		"""
		return self._get_waypoints(e)
	###
	# label(...)
	def _get_labels(self,n):
		"""
		Helper for label(...).  Returns the labels of nodes.
		"""
		if n == constants.NODES or n == constants.ALL:
			label_list = []
			
			for g in self.node2graphics.values():
				txt_id = g[1]
				txt = self.itemcget(txt_id,'text')
				
				label_list.append(txt)
					
			return txt_list
			
		elif type(n) is types.ListType:
			label_list = []
			
			for node in n:
				g = self.node2graphics[node]
				txt_id = g[1]
				txt = self.itemcget(txt_id,'text')
				
				label_list.append(txt)
					
			return label_list
		
		elif n in self.node2graphics:
			g = self.node2graphics[n]
			txt_id = g[1]
			return self.itemcget(txt_id,'text')
		
		else:
			raise Exception, 'Node(s) ' + str(n) + ' not found in graph' 
	
	def label(self,n,label=None):
		"""
		Return the label(s) of node(s) n.  n can be a single node, a list of nodes, NODES, or ALL.  The labels for the nodes
		is returned.  If multiple labels will be returned, a list of strings is returned.
		
		If label is specified, then the label of node n is set to the label specified.  If the label is '', then the 
		node's label is effectively removed.
		"""
		
		if label is None:
			return self._get_labels(n)
		else:
			if n not in self.node2graphics:
				raise Exception, 'Node ' + str(n) + ' not found in graph'

			g = self.node2graphics[n]
			self.itemconfig(g[1],text=label)
	###
	# pos(...)
	def move_node(self,n,x,y):
		"""
		Helper function for pos(...) and for drag_node behavior of UI.  Moves a node to a specified position.  This is
		intended for internal use only.  If you want to move a node, use pos(...).
		"""
		oval_id, txt_id = self.node2graphics[n]

		# move the oval
		node_radius = self.node2size[n]
		oval_coords = (x - node_radius,y - node_radius,x + node_radius, y + node_radius)
		self.coords(oval_id, oval_coords)
		
		# move the text
		self.coords(txt_id, x + node_radius, y - node_radius - TEXT_Y_BUFFER)
		
		# move the edges
		if self.graph.is_directed():
			elist = self.graph.E(n,directed=False)
		else:
			elist = self.graph.E(n)
			
		for e in elist:
			if e not in self.edge2graphics:
				continue
		
			einfo = self.edge2graphics[e]
			src_node = einfo[0]
			tgt_node = einfo[1]
			
			# If the selected node is the source
			if n == einfo[0]:
				line_idx = 0
			else: # then it's the target
				line_idx = -1

			line_id = einfo[2][line_idx]
			
			# if the selected node is the source
			if len(einfo[2]) > 1:
				line_coords = self.coords(line_id)
			
				src = (line_coords[0],line_coords[1])
				tgt = (line_coords[2],line_coords[3])
			
				if n == einfo[0]:
					line_coords[0] = x
					line_coords[1] = y
					
					src = (x,y)
					
					src = tkg.update_tgt_on_circle(tgt, src, self.node2size[src_node])
				else: # it's the target
					line_coords[2] = x
					line_coords[3] = y
	
					tgt = (x,y)
					
					tgt = tkg.update_tgt_on_circle(src, tgt, self.node2size[tgt_node])
				
				self.coords(line_id, (src[0],src[1],tgt[0],tgt[1]))
			else:
				if n == einfo[0]:
					self.coords(line_id, tkg.compute_line_coords_from_oval_info(oval_coords, self.coords(self.node2graphics[einfo[1]][0]), self.node2size[src_node], self.node2size[tgt_node]))
				else:
					self.coords(line_id, tkg.compute_line_coords_from_oval_info(self.coords(self.node2graphics[einfo[0]][0]),oval_coords, self.node2size[tgt_node], self.node2size[src_node]))

	def _get_pos(self,n):
		"""
		Helper function for pos(...).  Returns the positions of node(s) n.
		"""
		if n == constants.NODES or n == constants.ALL:
			pos_list = []
			
			for node,g in self.node2graphics.items():
				oval_id = g[0]
				coords = self.coords(oval_id)
				
				pos_list.append( (coords[0] + self.node2size[node], coords[1] + self.node2size[node]) )
					
			return pos_list
			
		elif type(n) is types.ListType:
			pos_list = []
			
			for node in n:
				g = self.node2graphics[node]
				oval_id = g[0]
				coords = self.coords(oval_id)
				
				pos_list.append( (coords[0] + self.node2size[node], coords[1] + self.node2size[node]) )
					
			return pos_list
		
		elif n in self.node2graphics:
			g = self.node2graphics[n]
			oval_id = g[0]
			coords = self.coords(oval_id)
			
			return (coords[0] + self.node2size[n], coords[1] + self.node2size[n])
		else:
			raise Exception, 'Node(s) ' + str(n) + ' not found in graph' 

	def pos(self,n,pos=None):
		"""
		Return the position of all nodes specified.  The input can be a single node, a list of nodes, NODES or ALL.  Positions
		of the indicated nodes are returned.
		
		If pos is specified, then only one node can be given as input.  This node's position will be set to the specified
		position.  The value of pos must be an x-y pair of values (either a tuple or a list).
		"""

		if pos is None:
			return self._get_pos(n)
			
		if n not in self.node2graphics:
			raise Exception, 'Node ' + str(n) + ' not found in graph'
		
		self.move_node(n,pos[0],pos[1])
	###
	# size(...)
	def _get_size(self,n):
		"""
		Helper method for size(...).  Returns the sizes of the nodes in n.
		"""
		if n == constants.NODES or n == constants.ALL:
			size_list = []
			
			for node in self.node2graphics:
				size_list.append( self.node2size[node] )
					
			return node_list
			
		elif type(n) is types.ListType:
			node_list = []
			
			for node in n:
				node_list.append( self.node2size[node] )
					
			return node_list
		
		elif n in self.node2graphics:
			return self.node2size[n]
		else:
			raise Exception, 'Node(s) ' + str(n) + ' not found in graph' 
		
	def _resize_node(self,n,size):
		"""
		Helper for size(...).  Resizes nodes and handles all the details.
		"""
		self.node2is_default[n][NODE_SIZE_IDX] = False
				
		old_size = self.node2size[n]
		self.node2size[n] = size
		
		g = self.node2graphics[n]
		oval_id, txt_id = g
		
		# resize the oval
		coords = self.coords(oval_id)
		x = coords[0] + old_size
		y = coords[1] + old_size
		
		self.coords(oval_id,(x - size, y - size, x + size, y + size))
		
		# reposition the text
		self.coords(txt_id,x + size, y - size - TEXT_Y_BUFFER)
		
		# resize all edges
		elist = None
		if self.graph.is_directed():
			elist = self.graph.E(n,directed=False)
		else:
			elist = self.graph.E(n)
			
		for e in elist:
			if e not in self.edge2graphics:
				continue
		
			einfo = self.edge2graphics[e]
			
			# If the selected node is the source
			if n == einfo[0]:
				line_idx = 0
			else: # then it's the target
				line_idx = -1

			line_id = einfo[2][line_idx]
			
			# if the selected node is the source
			if len(einfo[2]) > 1:
				line_coords = self.coords(line_id)
			
				src = (line_coords[0],line_coords[1])
				tgt = (line_coords[2],line_coords[3])
			
				if n == einfo[0]:
					line_coords[0] = x
					line_coords[1] = y
					
					src = (x,y)
					
					src = tkg.update_tgt_on_circle(tgt, src, size)
				else: # it's the target
					line_coords[2] = x
					line_coords[3] = y
	
					tgt = (x,y)
					
					tgt = tkg.update_tgt_on_circle(src, tgt, size)
				
				self.coords(line_id, (src[0],src[1],tgt[0],tgt[1]))
			else:
				if n == einfo[0]:
					self.coords(line_id, tkg.compute_line_coords_from_oval_info(oval_coords, self.coords(self.node2graphics[einfo[1]][0]), size))
				else:
					self.coords(line_id, tkg.compute_line_coords_from_oval_info(self.coords(self.node2graphics[einfo[0]][0]),oval_coords, size))

		
	def size(self,n,size=None):
		"""
		Returns the size of a node or collection of nodes.  n can be a single node or a list of nodes or NODES or ALL.
		The sizes of the nodes are returned.
		
		If size is specified, then the size of all nodes specified in n are set to the specified size.
		"""
		
		if size is None:
			return _get_size(n)
		
		if n == constants.NODES or n == constants.ALL:
			# set the size of all nodes
			widths = []
			for node in self.node2graphics:
				self._resize_node(node,size)
					
		elif type(n) is types.ListType:
			# set the width of all specified edges
			for node in n:
				self._resize_node(node,size)
		
		elif n in self.node2graphics:
			self._resize_node(n,size)
		else:
			raise Exception, 'Node(s) ' + str(n) + ' not found in graph'
		
class GraphFrame(Frame):
	"""
	The GraphFrame provides a fully functional Tkinter widget that renders and processes interactions
	with a PyWeaver graph.
	"""
	
	def __init__(self,root,graph,node_radius=None):
		Frame.__init__(self,root)
		
		self.gcanvas = GraphCanvas(self,graph,node_radius)
		
		self._build_ui()
		
	def _build_ui(self):
		
		# create scroll bars
		self.scrollX = Scrollbar(self, orient=HORIZONTAL)
		self.scrollY = Scrollbar(self, orient=VERTICAL)
		
		# link the scroll bars and canvas
		self.gcanvas['xscrollcommand'] = self.scrollX.set
		self.gcanvas['yscrollcommand'] = self.scrollY.set
		self.scrollX['command'] = self.gcanvas.xview
		self.scrollY['command'] = self.gcanvas.yview

		# layout
		self.scrollX.pack(side=BOTTOM, fill=X)
		self.scrollY.pack(side=RIGHT, fill=Y)
		self.gcanvas.pack(side=TOP,fill=BOTH, expand=1)
		
	####
	# All subsequent methods are pass-throughs to the GraphCanvas
	def default_node_color(self,c=None,reapply=False):
		"""
		Return the default color applied to nodes in the graph.
		
		If color is specified then the change is applied to all nodes that were not added with
		a color render hint.  If reapply is False, then the color of nodes already being 
		visualized are not changed.  If reapply is True, then the color of nodes whose color was
		not specified in a render hint or via a call to color(...) will be changed.
		"""
		if c is None:
			return self.gcanvas.default_node_color()
		else:
			self.gcanvas.default_node_color(c,reapply)

	def default_edge_color(self,c=None,reapply=False):
		"""
		Return the default color applied to edges in the graph.
		
		If color is specified then the change is applied to all edges added to the graph after
		the change.  If reapply is False, then any edges currently being visualized are not affected.
		If reapply is True, then any edges whose colors were not specified in a render hint or by a call
		to the color(...) method will have their color changed.
		"""	
		if c is None:
			return self.gcanvas.default_edge_color()
		else:
			self.gcanvas.default_edge_color(c,reapply)

	def color(self,x,color=None):
		"""
		If color is specified, set the color of the component(s) to the color indicated.
		
		If color is not specified, return the color(s) of the component(s).
		
		In both uses of this method, NODES, EDGES, and ALL can be used as components.
		"""
		if color is None:
			return self.gcanvas.color(x)
		else:
			self.gcanvas.color(x,color)

	def label(self,x,label=None):
		"""
		Return the label(s) of node(s) n.  n can be a single node, a list of nodes, NODES, or ALL.  The labels for the nodes
		is returned.  If multiple labels will be returned, a list of strings is returned.
		
		If label is specified, then the label of node n is set to the label specified.  If the label is '', then the 
		node's label is effectively removed.
		"""
		if label is None:
			return self.gcanvas.label(x)
		else:
			self.gcanvas.label(x,label)

	def pos(self,x,coords=None):
		"""
		Return the position of all nodes specified.  The input can be a single node, a list of nodes, NODES or ALL.  Positions
		of the indicated nodes are returned.
		
		If pos is specified, then only one node can be given as input.  This node's position will be set to the specified
		position.  The value of pos must be an x-y pair of values (either a tuple or a list).
		"""
		if coords is None:
			return self.gcanvas.coords(x)
		else:
			self.gcanvas.coords(x,coords)

	def size(self,x,size=None):
		"""
		Returns the size of a node or collection of nodes.  n can be a single node or a list of nodes or NODES or ALL.
		The sizes of the nodes are returned.
		
		If size is specified, then the size of all nodes specified in n are set to the specified size.
		"""
		if size is None:
			return self.gcanvas.size(x)
		else:
			self.gcanvas.size(x,size)

	def width(self,x,width=None):
		"""
		Set the width of the specified edge(s).  If no width is specified, then the width of the 
		specified edges is returned.  When e is EDGES or ALL, this method returns the width of all
		edges.
		"""
		if width is None:
			return self.gcanvas.width(x)
		else:
			self.gcanvas.width(x,width)

	def head(self,x,head=None):
		"""
		Set the head of the specified edge(s).  If no head is specified, then the head of the 
		specified edges is returned.  When e is EDGES or ALL, this method returns the head of all
		edges.
		"""
		if head is None:
			return self.gcanvas.head(x)
		else:
			self.gcanvas.head(x,head)

	def waypoints(self,e):
		"""
		The waypoints of the specified edges are returned.  A single edge, a list of edges, or EDGES or ALL are permitted 
		and this method returns the waypoints of the specified set of edges.
		"""
		self.gcanvas.waypoints(e)
		
	def add_selection_listener(self,sl):
		self.gcanvas.add_selection_listener(sl)
	
	def rm_selection_listener(self,sl):
		self.gcanvas.rm_selection_listener(sl)
		
	def selected(self):
		"""
		Return a set containing the currently selected nodes.  This set is guaranteed to be independent
		of the underlying implementation and, therefore, may be modified.
		"""
		return self.gcanvas.selected()
		
	def sync_graph(self,force_update=False):
		"""
		Write the current rendering properties of the graph in the canvas to the underlying graph's 
		render hints.
		"""
		self.gcanvas.sync_graph(force_update)

class GraphViewer:
	"""
	The GraphViewer class encapsulates the functionality of the GraphFrame into a windowed tool that can be easily launched.
	The easiest way to obtain a functional GraphViewer is to call the function view(...).  This will create a GraphViewer
	and launch the window in another thread.
	"""
	def __init__(self,root=None,graph=None):
		self.root = root
		
		if graph is None:
			self.graph = g_mod.Graph()
		else:
			self.graph = graph
		
		self.build_ui()
		
	def build_ui(self):
		
		# build a top level window if necessary
		if self.root is None:
			self.root = Toplevel()
			
		# configure the root
		self.root.title('Graph Viewer')
		
		# add a GraphFrame
		self.graph_frame = GraphFrame(self.root,self.graph)
		self.graph_frame.pack(fill='both',expand=True)


def inner_view(graph):
	root = Tk()

	gviewer = GraphViewer(root, graph)
		
	root.mainloop()
	
def view(graph):
	"""
	Launch an interactive visual interface for the graph specified.
	"""	
	
	thread.start_new_thread(inner_view,(graph,))

def test_GraphCanvas():

	from pyweaver.digraph import DiGraph

	class FooListener:
		def component_selected(self,component):
			print 'Selected: ', component
		def component_deselected(self,component):
			print 'Deselected: ', component
	
	class FooEdge:
		
		def __init__(self,ahead,width=1):
			self.ahead = ahead
			self.width = width
	
	graph = DiGraph()
	graph.add_edge("X","Y",FooEdge(">",1))
	graph.add_edge("X","Z",FooEdge('|',1))
	
	gv = view(graph, fork = True)
	gv.graph_frame.add_selection_listener(FooListener())
	print gv
		