"""
This modules provides functions that are used by the Tk graph visualization objects
and routines.
"""

"""
Copyright 2007 Derek Ruths

This file is part of PyWeaver.

PyWeaver is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PyWeaver is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with PyWeaver.  If not, see <http://www.gnu.org/licenses/>.
"""

import math

def compute_line_coords_from_oval_info(src_oval, tgt_oval, src_radius, tgt_radius):
	"""
	Compute the endpoints of the line connecting the boundaries of the circles specified.
	
	The format for specifying the a circle position is: (left_x, top_y, right_x, bottom_y)
	
	The return value is a 4-tuple: (x0, y0, x1, y1)
	"""
	src_x = src_oval[0] + src_radius 
	src_y = src_oval[1] + src_radius 
	tgt_x = tgt_oval[0] + tgt_radius 
	tgt_y = tgt_oval[1] + tgt_radius 

	return compute_line_coords_from_centers((src_x, src_y), (tgt_x, tgt_y), src_radius, tgt_radius)


def compute_line_coords_from_centers(src, tgt, src_radius, tgt_radius):	
	"""
	Compute the endpoints of the line connecting the boundaries of two circles.  The
	src and tgt specify the centers of both circles.  They are assumed to have the same radius.
	"""
	tgt = update_tgt_on_circle(src, tgt, tgt_radius)
	src = update_tgt_on_circle(tgt, src, src_radius)
	
	return (src[0],src[1],tgt[0],tgt[1])


def update_tgt_on_circle(src, tgt, radius):
	"""
	Update the location of the target of the arrow such that it lies on the boundary
	of the circle with radius specified.  It is assumed that the tgt position is the
	center of the circle.
	"""
	tgt_x = tgt[0]
	tgt_y = tgt[1]
	
	# compute the target location
	dx = src[0] - tgt_x
	dy = src[1] - tgt_y
	
	theta = math.atan2(dy,dx)
	
	tgt_x = tgt_x + math.cos(theta) * float(radius)
	tgt_y = tgt_y + math.sin(theta) * float(radius)
	
	return (tgt_x,tgt_y)
	
def compute_pt2line_distance(pt, line):
	"""
	Compute the distance between a point and a line segment.
	"""
	dx = float(line[2] - line[0])
	dy = float(line[3] - line[1])
	
	length2 = (dx * dx) + (dy * dy);
	
	u = ((float(pt[0] - line[0]) * dx) + (float(pt[1] - line[1]) * dy)) / length2;
	
	ix = float(line[0]) + (u * dx);
	iy = float(line[1]) + (u * dy);
	
	if u < 0:
		ix = line[0]
		iy = line[1]
	elif u > 1:
		ix = line[2]
		iy = line[3]
	
	dx = float(pt[0]) - ix
	dy = float(pt[1]) - iy
	
	dist = (dx * dx) + (dy * dy)
	
	return math.sqrt(dist)
